# Copyright 2015 Mirantis
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from zunclient import api_versions
from zunclient import exceptions
from zunclient.tests.unit import utils
from zunclient.v1 import versions


class APIVersionTestCase(utils.TestCase):
    def test_valid_version_strings(self):
        def _test_string(version, exp_major, exp_minor):
            v = api_versions.APIVersion(version)
            self.assertEqual(v.ver_major, exp_major)
            self.assertEqual(v.ver_minor, exp_minor)

        _test_string("1.1", 1, 1)
        _test_string("2.10", 2, 10)
        _test_string("5.234", 5, 234)
        _test_string("12.5", 12, 5)
        _test_string("2.0", 2, 0)
        _test_string("2.200", 2, 200)

    def test_null_version(self):
        v = api_versions.APIVersion()
        self.assertTrue(v.is_null())

    def test_invalid_version_strings(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "2")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "200")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "2.1.4")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "200.23.66.3")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "5 .3")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "5. 3")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "5.03")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "02.1")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "2.001")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, " 2.1")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, "2.1 ")

    def test_version_comparisons(self):
        v1 = api_versions.APIVersion("2.0")
        v2 = api_versions.APIVersion("2.5")
        v3 = api_versions.APIVersion("5.23")
        v4 = api_versions.APIVersion("2.0")
        v_null = api_versions.APIVersion()

        self.assertTrue(v1 < v2)
        self.assertTrue(v3 > v2)
        self.assertTrue(v1 != v2)
        self.assertTrue(v1 == v4)
        self.assertTrue(v1 != v_null)
        self.assertTrue(v_null == v_null)
        self.assertRaises(TypeError, v1.__le__, "2.1")

    def test_version_matches(self):
        v1 = api_versions.APIVersion("2.0")
        v2 = api_versions.APIVersion("2.5")
        v3 = api_versions.APIVersion("2.45")
        v4 = api_versions.APIVersion("3.3")
        v5 = api_versions.APIVersion("3.23")
        v6 = api_versions.APIVersion("2.0")
        v7 = api_versions.APIVersion("3.3")
        v8 = api_versions.APIVersion("4.0")
        v_null = api_versions.APIVersion()

        self.assertTrue(v2.matches(v1, v3))
        self.assertTrue(v2.matches(v1, v_null))
        self.assertTrue(v1.matches(v6, v2))
        self.assertTrue(v4.matches(v2, v7))
        self.assertTrue(v4.matches(v_null, v7))
        self.assertTrue(v4.matches(v_null, v8))
        self.assertFalse(v1.matches(v2, v3))
        self.assertFalse(v5.matches(v2, v4))
        self.assertFalse(v2.matches(v3, v1))

        self.assertRaises(ValueError, v_null.matches, v1, v3)

    def test_get_string(self):
        v1_string = "3.23"
        v1 = api_versions.APIVersion(v1_string)
        self.assertEqual(v1_string, v1.get_string())

        self.assertRaises(ValueError,
                          api_versions.APIVersion().get_string)


class UpdateHeadersTestCase(utils.TestCase):
    def test_api_version_is_null(self):
        headers = {}
        api_versions.update_headers(headers, api_versions.APIVersion())
        self.assertEqual({}, headers)

    def test_api_version_is_major(self):
        headers = {}
        api_versions.update_headers(headers, api_versions.APIVersion("7.0"))
        self.assertEqual({}, headers)

    def test_api_version_is_not_null(self):
        api_version = api_versions.APIVersion("2.3")
        headers = {}
        api_versions.update_headers(headers, api_version)
        self.assertEqual(
            {"OpenStack-API-Version":
             "container %s" % api_version.get_string()},
            headers)


class GetAPIVersionTestCase(utils.TestCase):
    def test_get_available_client_versions(self):
        output = api_versions.get_available_major_versions()
        self.assertNotEqual([], output)

    def test_wrong_format(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.get_api_version, "something_wrong")

    @mock.patch("zunclient.api_versions.APIVersion")
    def test_only_major_part_is_presented(self, mock_apiversion):
        version = 7
        self.assertEqual(mock_apiversion.return_value,
                         api_versions.get_api_version(version))
        mock_apiversion.assert_called_once_with("%s.latest" % str(version))

    @mock.patch("zunclient.api_versions.APIVersion")
    def test_major_and_minor_parts_is_presented(self, mock_apiversion):
        version = "2.7"
        self.assertEqual(mock_apiversion.return_value,
                         api_versions.get_api_version(version))
        mock_apiversion.assert_called_once_with(version)


class WrapsTestCase(utils.TestCase):

    def _get_obj_with_vers(self, vers):
        return mock.MagicMock(api_version=api_versions.APIVersion(vers))

    def _side_effect_of_vers_method(self, *args, **kwargs):
        m = mock.MagicMock(start_version=args[1], end_version=args[2])
        m.name = args[0]
        return m

    @mock.patch("zunclient.api_versions._get_function_name")
    @mock.patch("zunclient.api_versions.VersionedMethod")
    def test_end_version_is_none(self, mock_versioned_method, mock_name):
        func_name = "foo"
        mock_name.return_value = func_name
        mock_versioned_method.side_effect = self._side_effect_of_vers_method

        @api_versions.wraps("2.2")
        def foo(*args, **kwargs):
            pass

        foo(self._get_obj_with_vers("2.4"))

        mock_versioned_method.assert_called_once_with(
            func_name, api_versions.APIVersion("2.2"),
            api_versions.APIVersion("2.latest"), mock.ANY)

    @mock.patch("zunclient.api_versions._get_function_name")
    @mock.patch("zunclient.api_versions.VersionedMethod")
    def test_start_and_end_version_are_presented(self, mock_versioned_method,
                                                 mock_name):
        func_name = "foo"
        mock_name.return_value = func_name
        mock_versioned_method.side_effect = self._side_effect_of_vers_method

        @api_versions.wraps("2.2", "2.6")
        def foo(*args, **kwargs):
            pass

        foo(self._get_obj_with_vers("2.4"))

        mock_versioned_method.assert_called_once_with(
            func_name, api_versions.APIVersion("2.2"),
            api_versions.APIVersion("2.6"), mock.ANY)

    @mock.patch("zunclient.api_versions._get_function_name")
    @mock.patch("zunclient.api_versions.VersionedMethod")
    def test_api_version_doesnt_match(self, mock_versioned_method, mock_name):
        func_name = "foo"
        mock_name.return_value = func_name
        mock_versioned_method.side_effect = self._side_effect_of_vers_method

        @api_versions.wraps("2.2", "2.6")
        def foo(*args, **kwargs):
            pass

        self.assertRaises(exceptions.VersionNotFoundForAPIMethod,
                          foo, self._get_obj_with_vers("2.1"))

        mock_versioned_method.assert_called_once_with(
            func_name, api_versions.APIVersion("2.2"),
            api_versions.APIVersion("2.6"), mock.ANY)

    def test_define_method_is_actually_called(self):
        checker = mock.MagicMock()

        @api_versions.wraps("2.2", "2.6")
        def some_func(*args, **kwargs):
            checker(*args, **kwargs)

        obj = self._get_obj_with_vers("2.4")
        some_args = ("arg_1", "arg_2")
        some_kwargs = {"key1": "value1", "key2": "value2"}

        some_func(obj, *some_args, **some_kwargs)

        checker.assert_called_once_with(*((obj,) + some_args), **some_kwargs)


class DiscoverVersionTestCase(utils.TestCase):
    def setUp(self):
        super(DiscoverVersionTestCase, self).setUp()
        self.orig_max = api_versions.MAX_API_VERSION
        self.orig_min = api_versions.MIN_API_VERSION
        self.addCleanup(self._clear_fake_version)

    def _clear_fake_version(self):
        api_versions.MAX_API_VERSION = self.orig_max
        api_versions.MIN_API_VERSION = self.orig_min

    def test_server_is_too_new(self):
        fake_client = mock.MagicMock()
        fake_client.versions.get_current.return_value = mock.MagicMock(
            max_version="1.7", min_version="1.4")
        api_versions.MAX_API_VERSION = "1.3"
        api_versions.MIN_API_VERSION = "1.1"
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.discover_version, fake_client,
                          api_versions.APIVersion('1.latest'))

    def test_server_is_too_old(self):
        fake_client = mock.MagicMock()
        fake_client.versions.get_current.return_value = mock.MagicMock(
            max_version="1.7", min_version="1.4")
        api_versions.MAX_API_VERSION = "1.10"
        api_versions.MIN_API_VERSION = "1.9"

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.discover_version, fake_client,
                          api_versions.APIVersion('1.latest'))

    def test_server_end_version_is_the_latest_one(self):
        fake_client = mock.MagicMock()
        fake_client.versions.get_current.return_value = mock.MagicMock(
            max_version="1.7", min_version="1.4")
        api_versions.MAX_API_VERSION = "1.11"
        api_versions.MIN_API_VERSION = "1.1"

        self.assertEqual(
            "1.7",
            api_versions.discover_version(
                fake_client,
                api_versions.APIVersion('1.latest')).get_string())

    def test_client_end_version_is_the_latest_one(self):
        fake_client = mock.MagicMock()
        fake_client.versions.get_current.return_value = mock.MagicMock(
            max_version="1.16", min_version="1.4")
        api_versions.MAX_API_VERSION = "1.11"
        api_versions.MIN_API_VERSION = "1.1"

        self.assertEqual(
            "1.11",
            api_versions.discover_version(
                fake_client,
                api_versions.APIVersion('1.latest')).get_string())

    def test_server_without_microversion(self):
        fake_client = mock.MagicMock()
        fake_client.versions.get_current.return_value = mock.MagicMock(
            max_version='', min_version='')
        api_versions.MAX_API_VERSION = "1.11"
        api_versions.MIN_API_VERSION = "1.1"

        self.assertEqual(
            "1.1",
            api_versions.discover_version(
                fake_client,
                api_versions.APIVersion('1.latest')).get_string())

    def test_server_without_microversion_and_no_version_field(self):
        fake_client = mock.MagicMock()
        fake_client.versions.get_current.return_value = versions.Version(
            None, {})
        api_versions.MAX_API_VERSION = "1.11"
        api_versions.MIN_API_VERSION = "1.1"

        self.assertEqual(
            "1.1",
            api_versions.discover_version(
                fake_client,
                api_versions.APIVersion('1.latest')).get_string())



#    Copyright 2017 Arm Limited.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from oslo_serialization import jsonutils

from zunclient.common import cliutils as utils
from zunclient.common import template_utils
from zunclient.common import utils as zun_utils
from zunclient.i18n import _


def _show_capsule(capsule):
    zun_utils.format_container_addresses(capsule)
    utils.print_dict(capsule._info)


@utils.arg('-f', '--template-file', metavar='<file>',
           required=True, help=_('Path to the template.'))
def do_capsule_create(cs, args):
    """Create a capsule."""
    opts = {}
    if args.template_file:
        template = template_utils.get_template_contents(
            args.template_file)
        opts['template'] = template
        cs.capsules.create(**opts)
        print("Request to create capsule has been accepted.")


@utils.arg('--all-projects',
           action="store_true",
           default=False,
           help='List containers in all projects')
@utils.arg('--marker',
           metavar='<marker>',
           default=None,
           help='The last container UUID of the previous page; '
                'displays list of containers after "marker".')
@utils.arg('--limit',
           metavar='<limit>',
           type=int,
           help='Maximum number of containers to return')
@utils.arg('--sort-key',
           metavar='<sort-key>',
           help='Column to sort results by')
@utils.arg('--sort-dir',
           metavar='<sort-dir>',
           choices=['desc', 'asc'],
           help='Direction to sort. "asc" or "desc".')
def do_capsule_list(cs, args):
    """Print a list of available capsules."""
    opts = {}
    opts['all_projects'] = args.all_projects
    opts['marker'] = args.marker
    opts['limit'] = args.limit
    opts['sort_key'] = args.sort_key
    opts['sort_dir'] = args.sort_dir
    opts = zun_utils.remove_null_parms(**opts)
    capsules = cs.capsules.list(**opts)
    zun_utils.list_capsules(capsules)


@utils.arg('capsules',
           metavar='<capsule>',
           nargs='+',
           help='ID or name of the (capsule)s to delete.')
def do_capsule_delete(cs, args):
    """Delete specified capsules."""
    for capsule in args.capsules:
        try:
            cs.capsules.delete(capsule)
            print("Request to delete capsule %s has been accepted." %
                  capsule)
        except Exception as e:
            print("Delete for capsule %(capsule)s failed: %(e)s" %
                  {'capsule': capsule, 'e': e})


@utils.arg('capsule',
           metavar='<capsule>',
           help='ID or name of the capsule to show.')
@utils.arg('-f', '--format',
           metavar='<format>',
           action='store',
           choices=['json', 'yaml', 'table'],
           default='table',
           help='Print representation of the capsule. '
                'The choices of the output format is json,table,yaml. '
                'Defaults to table. ')
def do_capsule_describe(cs, args):
    """Show details of a capsule."""
    capsule = cs.capsules.describe(args.capsule)
    if args.format == 'json':
        print(jsonutils.dumps(capsule._info, indent=4, sort_keys=True))
    elif args.format == 'yaml':
        print(yaml.safe_dump(capsule._info, default_flow_style=False))
    elif args.format == 'table':
        _show_capsule(capsule)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zunclient import api_versions
from zunclient.common import cliutils as utils


def do_version_list(cs, args):
    """List all API versions."""
    print("Client supported API versions:")
    print("Minimum version %(v)s" %
          {'v': api_versions.MIN_API_VERSION})
    print("Maximum version %(v)s" %
          {'v': api_versions.MAX_API_VERSION})

    print("\nServer supported API versions:")
    result = cs.versions.list()
    columns = ["Id", "Status", "Min Version", "Max Version"]
    utils.print_list(result, columns)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from osc_lib.command import command
from osc_lib import utils


def _get_client(obj, parsed_args):
    obj.log.debug("take_action(%s)" % parsed_args)
    return obj.app.client_manager.container


class ListService(command.Lister):
    """Print a list of zun services."""

    log = logging.getLogger(__name__ + ".ListService")

    def get_parser(self, prog_name):
        parser = super(ListService, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        client = _get_client(self, parsed_args)
        services = client.services.list()
        columns = ('Id', 'Host', 'Binary', 'State', 'Disabled',
                   'Disabled Reason', 'Updated At',
                   'Availability Zone')
        return (columns, (utils.get_item_properties(service, columns)
                          for service in services))


class DeleteService(command.Command):
    """Delete the Zun binaries/services."""

    log = logging.getLogger(__name__ + ".DeleteService")

    def get_parser(self, prog_name):
        parser = super(DeleteService, self).get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            help='Name of host')
        parser.add_argument(
            'binary',
            metavar='<binary>',
            help='Name of the binary to delete')
        return parser

    def take_action(self, parsed_args):
        client = _get_client(self, parsed_args)
        host = parsed_args.host
        binary = parsed_args.binary
        try:
            client.services.delete(host, binary)
            print("Request to delete binary %s on host %s has been accepted." %
                  (binary, host))
        except Exception as e:
            print("Delete for binary %s on host %s failed: %s" %
                  (binary, host, e))


class EnableService(command.ShowOne):
    """Enable the Zun service."""
    log = logging.getLogger(__name__ + ".EnableService")

    def get_parser(self, prog_name):
        parser = super(EnableService, self).get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            help='Name of host')
        parser.add_argument(
            'binary',
            metavar='<binary>',
            help='Name of the binary to enable')
        return parser

    def take_action(self, parsed_args):
        client = _get_client(self, parsed_args)
        host = parsed_args.host
        binary = parsed_args.binary
        res = client.services.enable(host, binary)
        columns = ('Host', 'Binary', 'Disabled', 'Disabled Reason')
        return columns, (utils.get_dict_properties(res[1]['service'],
                                                   columns))


class DisableService(command.ShowOne):
    """Disable the Zun service."""
    log = logging.getLogger(__name__ + ".DisableService")

    def get_parser(self, prog_name):
        parser = super(DisableService, self).get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            help='Name of host')
        parser.add_argument(
            'binary',
            metavar='<binary>',
            help='Name of the binary to disable')
        parser.add_argument(
            '--reason',
            metavar='<reason>',
            help='Reason for disabling service')
        return parser

    def take_action(self, parsed_args):
        client = _get_client(self, parsed_args)
        host = parsed_args.host
        binary = parsed_args.binary
        reason = parsed_args.reason
        res = client.services.disable(host, binary, reason)
        columns = ('Host', 'Binary', 'Disabled', 'Disabled Reason')
        return columns, (utils.get_dict_properties(res[1]['service'],
                                                   columns))


class ForceDownService(command.ShowOne):
    """Force the Zun service to down or up."""
    log = logging.getLogger(__name__ + ".ForceDownService")

    def get_parser(self, prog_name):
        parser = super(ForceDownService, self).get_parser(prog_name)
        parser.add_argument(
            'host',
            metavar='<host>',
            help='Name of host')
        parser.add_argument(
            'binary',
            metavar='<binary>',
            help='Name of the binary to disable')
        parser.add_argument(
            '--unset',
            dest='force_down',
            help='Unset the force state down of service',
            action='store_false',
            default=True)
        return parser

    def take_action(self, parsed_args):
        client = _get_client(self, parsed_args)
        host = parsed_args.host
        binary = parsed_args.binary
        force_down = parsed_args.force_down
        res = client.services.force_down(host, binary, force_down)
        columns = ('Host', 'Binary', 'Forced_down')
        return columns, (utils.get_dict_properties(res[1]['service'],
                                                   columns))



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1 import loading
from keystoneauth1 import session as ksa_session

from zunclient.common import httpclient
from zunclient.v1 import actions
from zunclient.v1 import availability_zones as az
from zunclient.v1 import capsules
from zunclient.v1 import containers
from zunclient.v1 import hosts
from zunclient.v1 import images
from zunclient.v1 import quota_classes
from zunclient.v1 import quotas
from zunclient.v1 import registries
from zunclient.v1 import services
from zunclient.v1 import versions


class Client(object):
    """Top-level object to access the OpenStack Container API."""

    def __init__(self, api_version=None, auth_token=None,
                 auth_type='password', auth_url=None, endpoint_override=None,
                 interface='public', insecure=False, password=None,
                 project_domain_id=None, project_domain_name=None,
                 project_id=None, project_name=None, region_name=None,
                 service_name=None, service_type='container', session=None,
                 user_domain_id=None, user_domain_name=None,
                 username=None, cacert=None, cert=None, key=None, **kwargs):
        """Initialization of Client object.

        :param api_version: Container API version
        :type api_version: zunclient.api_version.APIVersion
        :param str auth_token: Auth token
        :param str auth_url: Auth URL
        :param str auth_type: Auth Type
        :param str endpoint_override: Bypass URL
        :param str interface: Interface
        :param str insecure: Allow insecure
        :param str password: User password
        :param str project_domain_id: ID of project domain
        :param str project_domain_name: Nam of project domain
        :param str project_id: Project/Tenant ID
        :param str project_name: Project/Tenant Name
        :param str region_name: Region Name
        :param str service_name: Service Name
        :param str service_type: Service Type
        :param str session: Session
        :param str user_domain_id: ID of user domain
        :param str user_id: User ID
        :param str username: Username
        :param str cacert: CA certificate
        """
        if endpoint_override and auth_token:
            auth_type = 'admin_token'
            session = None
            loader_kwargs = {
                'token': auth_token,
                'endpoint': endpoint_override
            }
        elif auth_token and not session:
            auth_type = 'token'
            loader_kwargs = {
                'token': auth_token,
                'auth_url': auth_url,
                'project_domain_id': project_domain_id,
                'project_domain_name': project_domain_name,
                'project_id': project_id,
                'project_name': project_name,
                'user_domain_id': user_domain_id,
                'user_domain_name': user_domain_name
            }
        else:
            loader_kwargs = {
                'auth_url': auth_url,
                'password': password,
                'project_domain_id': project_domain_id,
                'project_domain_name': project_domain_name,
                'project_id': project_id,
                'project_name': project_name,
                'user_domain_id': user_domain_id,
                'user_domain_name': user_domain_name,
                'username': username,
            }

        # Backwards compatibility for people not passing in Session
        if session is None:
            loader = loading.get_plugin_loader(auth_type)
            # This should be able to handle v2 and v3 Keystone Auth
            auth_plugin = loader.load_from_options(**loader_kwargs)
            if cert and key:
                cert = cert, key
            session = ksa_session.Session(auth=auth_plugin,
                                          verify=(cacert or not insecure),
                                          cert=cert)
        client_kwargs = {}
        if not endpoint_override:
            try:
                # Trigger an auth error so that we can throw the exception
                # we always have
                session.get_endpoint(
                    service_name=service_name,
                    service_type=service_type,
                    interface=interface,
                    region_name=region_name
                )
            except Exception:
                raise RuntimeError('Not authorized')
        else:
            client_kwargs = {'endpoint_override': endpoint_override}

        self.http_client = httpclient.SessionClient(service_type=service_type,
                                                    service_name=service_name,
                                                    interface=interface,
                                                    region_name=region_name,
                                                    session=session,
                                                    api_version=api_version,
                                                    **client_kwargs)
        self.containers = containers.ContainerManager(self.http_client)
        self.images = images.ImageManager(self.http_client)
        self.services = services.ServiceManager(self.http_client)
        self.hosts = hosts.HostManager(self.http_client)
        self.versions = versions.VersionManager(self.http_client)
        self.capsules = capsules.CapsuleManager(self.http_client)
        self.availability_zones = az.AvailabilityZoneManager(self.http_client)
        self.actions = actions.ActionManager(self.http_client)
        self.quotas = quotas.QuotaManager(self.http_client)
        self.quota_classes = quota_classes.QuotaClassManager(self.http_client)
        self.registries = registries.RegistryManager(self.http_client)

    @property
    def api_version(self):
        return self.http_client.api_version



# Copyright 2010 Jacob Kaplan-Moss
# Copyright 2011 OpenStack Foundation
# Copyright 2012 Grid Dynamics
# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import fcntl
import os
from oslo_log import log as logging
import select
import signal
import socket
import ssl
import struct
import sys
import termios
import time
import tty
from urllib import parse as urlparse
import websocket

from zunclient.common.apiclient import exceptions as acexceptions
from zunclient.common.websocketclient import exceptions

LOG = logging.getLogger(__name__)

DEFAULT_API_VERSION = '1'
DEFAULT_ENDPOINT_TYPE = 'publicURL'
DEFAULT_SERVICE_TYPE = 'container'


class BaseClient(object):

    def __init__(self, zunclient, url, id, escape='~',
                 close_wait=0.5):
        self.url = url
        self.id = id
        self.escape = escape
        self.close_wait = close_wait
        self.cs = zunclient

    def connect(self):
        raise NotImplementedError()

    def fileno(self):
        raise NotImplementedError()

    def send(self, data):
        raise NotImplementedError()

    def recv(self):
        raise NotImplementedError()

    def tty_resize(self, height, width):
        """Resize the tty session

        Get the client and send the tty size data to zun api server
        The environment variables need to get when implement sending
        operation.
        """
        raise NotImplementedError()

    def start_loop(self):
        self.poll = select.poll()
        self.poll.register(sys.stdin,
                           select.POLLIN | select.POLLHUP
                           | select.POLLPRI | select.POLLNVAL)
        self.poll.register(self.fileno(),
                           select.POLLIN | select.POLLHUP |
                           select.POLLPRI | select.POLLNVAL)

        self.start_of_line = False
        self.read_escape = False
        with WINCHHandler(self):
            try:
                self.setup_tty()
                self.run_forever()
            except socket.error as e:
                raise exceptions.ConnectionFailed(e)
            except websocket.WebSocketConnectionClosedException as e:
                raise exceptions.Disconnected(e)
            finally:
                self.restore_tty()

    def run_forever(self):
        LOG.debug('starting main loop in client')
        self.quit = False
        quitting = False
        when = None

        while True:
            try:
                for fd, event in self.poll.poll(500):
                    if fd == self.fileno():
                        self.handle_socket(event)
                    elif fd == sys.stdin.fileno():
                        self.handle_stdin(event)
            except select.error as e:
                # POSIX signals interrupt select()
                no = e.errno
                if no == errno.EINTR:
                    continue
                else:
                    raise e

            if self.quit and not quitting:
                LOG.debug('entering close_wait')
                quitting = True
                when = time.time() + self.close_wait

            if quitting and time.time() > when:
                LOG.debug('quitting')
                break

    def setup_tty(self):
        if os.isatty(sys.stdin.fileno()):
            LOG.debug('putting tty into raw mode')
            self.old_settings = termios.tcgetattr(sys.stdin)
            tty.setraw(sys.stdin)

    def restore_tty(self):
        if os.isatty(sys.stdin.fileno()):
            LOG.debug('restoring tty configuration')
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN,
                              self.old_settings)

    def handle_stdin(self, event):
        if event in (select.POLLHUP, select.POLLNVAL):
            LOG.debug('event %d on stdin', event)

            LOG.debug('eof on stdin')
            self.poll.unregister(sys.stdin)
            self.quit = True

        data = os.read(sys.stdin.fileno(), 1024)

        if not data:
            return

        if self.start_of_line and data == self.escape:
            self.read_escape = True
            return

        if self.read_escape and data == '.':
            LOG.debug('exit by local escape code')
            raise exceptions.UserExit()
        elif self.read_escape:
            self.read_escape = False
            self.send(self.escape)

        self.send(data)

        if data == '\r':
            self.start_of_line = True
        else:
            self.start_of_line = False

    def handle_socket(self, event):
        if event in (select.POLLHUP, select.POLLNVAL):
            self.poll.unregister(self.fileno())
            self.quit = True

        data = self.recv()
        if not data:
            self.poll.unregister(self.fileno())
            self.quit = True
            return

        if isinstance(data, bytes):
            data = data.decode("utf-8")
        sys.stdout.write(data)
        sys.stdout.flush()

    def handle_resize(self):
        """send the POST to resize the tty session size in container.

        Resize the container's PTY.
        If `size` is not None, it must be a tuple of (height,width), otherwise
        it will be determined by the size of the current TTY.
        """
        size = self.tty_size(sys.stdout)

        if size is not None:
            rows, cols = size
            try:
                self.tty_resize(height=rows, width=cols)
            except IOError:  # Container already exited
                pass
            except acexceptions.BadRequest:
                pass

    def tty_size(self, fd):
        """Get the tty size

        Return a tuple (rows,cols) representing the size of the TTY `fd`.

        The provided file descriptor should be the stdout stream of the TTY.

        If the TTY size cannot be determined, returns None.
        """

        if not os.isatty(fd.fileno()):
            return None

        try:
            dims = struct.unpack('hh', fcntl.ioctl(fd,
                                                   termios.TIOCGWINSZ,
                                                   'hhhh'))
        except Exception:
            try:
                dims = (os.environ['LINES'], os.environ['COLUMNS'])
            except Exception:
                return None

        return dims


class WebSocketClient(BaseClient):

    def __init__(self, zunclient, url, id, escape='~',
                 close_wait=0.5):
        super(WebSocketClient, self).__init__(
            zunclient, url, id, escape, close_wait)

    def connect(self):
        url = self.url
        LOG.debug('connecting to: %s', url)
        try:
            self.ws = websocket.create_connection(
                url, skip_utf8_validation=True,
                origin=self._compute_origin_header(url),
                sslopt={'cert_reqs': ssl.CERT_REQUIRED,
                        'ca_certs': self.get_system_ca_file()},
                subprotocols=["binary", "base64"])
            print('connected to %s, press Enter to continue' % self.id)
            print('type %s. to disconnect' % self.escape)
        except socket.error as e:
            raise exceptions.ConnectionFailed(e)
        except websocket.WebSocketConnectionClosedException as e:
            raise exceptions.ConnectionFailed(e)
        except websocket.WebSocketBadStatusException as e:
            raise exceptions.ConnectionFailed(e)

    def _compute_origin_header(self, url):
        origin = urlparse.urlparse(url)
        if origin.scheme == 'wss':
            return "https://%s:%s" % (origin.hostname, origin.port)
        else:
            return "http://%s:%s" % (origin.hostname, origin.port)

    def fileno(self):
        return self.ws.fileno()

    def send(self, data):
        self.ws.send_binary(data)

    def recv(self):
        return self.ws.recv()

    @staticmethod
    def get_system_ca_file():
        """Return path to system default CA file."""
        # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
        # Suse, FreeBSD/OpenBSD
        ca_path = ['/etc/ssl/certs/ca-certificates.crt',
                   '/etc/pki/tls/certs/ca-bundle.crt',
                   '/etc/ssl/ca-bundle.pem',
                   '/etc/ssl/cert.pem']
        for ca in ca_path:
            if os.path.exists(ca):
                return ca
        return None


class AttachClient(WebSocketClient):

    def tty_resize(self, height, width):
        """Resize the tty session

        Get the client and send the tty size data to zun api server
        The environment variables need to get when implement sending
        operation.
        """
        height = str(height)
        width = str(width)

        self.cs.containers.resize(self.id, width, height)


class ExecClient(WebSocketClient):

    def __init__(self, zunclient, url, exec_id, id, escape='~',
                 close_wait=0.5):
        super(ExecClient, self).__init__(zunclient, url, id, escape,
                                         close_wait)
        self.exec_id = exec_id

    def tty_resize(self, height, width):
        """Resize the tty session

        Get the client and send the tty size data to zun api server
        The environment variables need to get when implement sending
        operation.
        """
        height = str(height)
        width = str(width)

        self.cs.containers.execute_resize(self.id, self.exec_id, width, height)


class WINCHHandler(object):
    """WINCH Signal handler

    WINCH Signal handler to keep the PTY correctly sized.
    """

    def __init__(self, client):
        """Initialize a new WINCH handler for the given PTY.

        Initializing a handler has no immediate side-effects. The `start()`
        method must be invoked for the signals to be trapped.
        """

        self.client = client
        self.original_handler = None

    def __enter__(self):
        """Enter

        Invoked on entering a `with` block.
        """

        self.start()
        return self

    def __exit__(self, *_):
        """Exit

        Invoked on exiting a `with` block.
        """

        self.stop()

    def start(self):
        """Start

        Start trapping WINCH signals and resizing the PTY.
        This method saves the previous WINCH handler so it can be restored on
        `stop()`.
        """

        def handle(signum, frame):
            if signum == signal.SIGWINCH:
                LOG.debug("Send command to resize the tty session")
                self.client.handle_resize()

        self.original_handler = signal.signal(signal.SIGWINCH, handle)

    def stop(self):
        """stop

        Stop trapping WINCH signals and restore the previous WINCH handler.
        """

        if self.original_handler is not None:
            signal.signal(signal.SIGWINCH, self.original_handler)


def do_attach(zunclient, url, container_id, escape, close_wait):
    if url.startswith("ws://") or url.startswith("wss://"):
        try:
            wscls = AttachClient(zunclient=zunclient, url=url,
                                 id=container_id, escape=escape,
                                 close_wait=close_wait)
            wscls.connect()
            wscls.handle_resize()
            wscls.start_loop()
        except exceptions.ContainerWebSocketException as e:
            print("%(e)s:%(container)s" %
                  {'e': e, 'container': container_id})
    else:
        raise exceptions.InvalidWebSocketLink(container_id)


def do_exec(zunclient, url, container_id, exec_id, escape, close_wait):
    if url.startswith("ws://") or url.startswith("wss://"):
        try:
            wscls = ExecClient(zunclient=zunclient, url=url,
                               exec_id=exec_id,
                               id=container_id, escape=escape,
                               close_wait=close_wait)
            wscls.connect()
            wscls.handle_resize()
            wscls.start_loop()
        except exceptions.ContainerWebSocketException as e:
            print("%(e)s:%(container)s" %
                  {'e': e, 'container': container_id})
    else:
        raise exceptions.InvalidWebSocketLink(container_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import testtools
from testtools import matchers
from zunclient import exceptions
from zunclient.tests.unit import utils
from zunclient.v1 import registries

REGISTRY1 = {'uuid': 'cb9d9848-6c12-411b-a915-6eadc7bd2871',
             'name': 'test1',
             'domain': 'testdomain.io',
             'username': 'testuser1',
             'password': 'testpassword1',
             }

REGISTRY2 = {'uuid': 'c2d923ef-e9b5-4e6b-a8aa-7c54a715b370',
             'name': 'test2',
             'domain': 'otherdomain.io',
             'username': 'testuser2',
             'password': 'testpassword2',
             }


CREATE_REGISTRY1 = copy.deepcopy(REGISTRY1)
del CREATE_REGISTRY1['uuid']

UPDATE_REGISTRY1 = copy.deepcopy(REGISTRY1)
del UPDATE_REGISTRY1['uuid']
UPDATE_REGISTRY1['name'] = 'newname'

fake_responses = {
    '/v1/registries':
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
        'POST': (
            {},
            {'registry': CREATE_REGISTRY1},
        ),
    },
    '/v1/registries/?limit=2':
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
    },
    '/v1/registries/?marker=%s' % REGISTRY2['uuid']:
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
    },
    '/v1/registries/?limit=2&marker=%s' % REGISTRY2['uuid']:
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
    },
    '/v1/registries/?sort_dir=asc':
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
    },
    '/v1/registries/?sort_key=uuid':
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
    },
    '/v1/registries/?sort_key=uuid&sort_dir=desc':
    {
        'GET': (
            {},
            {'registries': [REGISTRY1, REGISTRY2]},
        ),
    },
    '/v1/registries/%s' % REGISTRY1['uuid']:
    {
        'GET': (
            {},
            {'registry': REGISTRY1}
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            {'registry': UPDATE_REGISTRY1},
        ),
    },
}


class RegistryManagerTest(testtools.TestCase):

    def setUp(self):
        super(RegistryManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = registries.RegistryManager(self.api)

    def test_registry_create(self):
        registries = self.mgr.create(**CREATE_REGISTRY1)
        expect = [
            ('POST', '/v1/registries', {}, {'registry': CREATE_REGISTRY1})
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(registries)

    def test_registry_create_fail(self):
        create_registry_fail = copy.deepcopy(CREATE_REGISTRY1)
        create_registry_fail["wrong_key"] = "wrong"
        self.assertRaisesRegex(exceptions.InvalidAttribute,
                               ("Key must be in %s" %
                                ','.join(registries.CREATION_ATTRIBUTES)),
                               self.mgr.create, **create_registry_fail)
        self.assertEqual([], self.api.calls)

    def test_registries_list(self):
        registries = self.mgr.list()
        expect = [
            ('GET', '/v1/registries', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(registries, matchers.HasLength(2))

    def _test_registries_list_with_filters(self, limit=None, marker=None,
                                           sort_key=None, sort_dir=None,
                                           expect=[]):
        registries_filter = self.mgr.list(limit=limit, marker=marker,
                                          sort_key=sort_key,
                                          sort_dir=sort_dir)
        self.assertEqual(expect, self.api.calls)
        self.assertThat(registries_filter, matchers.HasLength(2))

    def test_registries_list_with_limit(self):
        expect = [
            ('GET', '/v1/registries/?limit=2', {}, None),
        ]
        self._test_registries_list_with_filters(
            limit=2,
            expect=expect)

    def test_registries_list_with_marker(self):
        expect = [
            ('GET', '/v1/registries/?marker=%s' % REGISTRY2['uuid'],
             {}, None),
        ]
        self._test_registries_list_with_filters(
            marker=REGISTRY2['uuid'],
            expect=expect)

    def test_registries_list_with_marker_limit(self):
        expect = [
            ('GET', '/v1/registries/?limit=2&marker=%s' % REGISTRY2['uuid'],
             {}, None),
        ]
        self._test_registries_list_with_filters(
            limit=2, marker=REGISTRY2['uuid'],
            expect=expect)

    def test_coontainer_list_with_sort_dir(self):
        expect = [
            ('GET', '/v1/registries/?sort_dir=asc', {}, None),
        ]
        self._test_registries_list_with_filters(
            sort_dir='asc',
            expect=expect)

    def test_registry_list_with_sort_key(self):
        expect = [
            ('GET', '/v1/registries/?sort_key=uuid', {}, None),
        ]
        self._test_registries_list_with_filters(
            sort_key='uuid',
            expect=expect)

    def test_registry_list_with_sort_key_dir(self):
        expect = [
            ('GET', '/v1/registries/?sort_key=uuid&sort_dir=desc', {}, None),
        ]
        self._test_registries_list_with_filters(
            sort_key='uuid', sort_dir='desc',
            expect=expect)

    def test_registry_show(self):
        registry = self.mgr.get(REGISTRY1['uuid'])
        expect = [
            ('GET', '/v1/registries/%s' % REGISTRY1['uuid'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(REGISTRY1['name'], registry._info['registry']['name'])
        self.assertEqual(REGISTRY1['uuid'], registry._info['registry']['uuid'])

    def test_registry_update(self):
        registry = self.mgr.update(REGISTRY1['uuid'], **UPDATE_REGISTRY1)
        expect = [
            ('PATCH', '/v1/registries/%s' % REGISTRY1['uuid'],
             {},
             {'registry': UPDATE_REGISTRY1})
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(UPDATE_REGISTRY1['name'],
                         registry._info['registry']['name'])

    def test_registries_delete(self):
        registries = self.mgr.delete(REGISTRY1['uuid'])
        expect = [
            ('DELETE', '/v1/registries/%s' % (REGISTRY1['uuid']),
             {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(registries)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from testtools import matchers
from zunclient.tests.unit import utils
from zunclient.v1 import availability_zones as az

AZ1 = {'availability_zone': 'fake-az-1'}

AZ2 = {'availability_zone': 'fake-az-2'}

fake_responses = {
    '/v1/availability_zones':
    {
        'GET': (
            {},
            {'availability_zones': [AZ1, AZ2]},
        ),
    },
}


class AZManagerTest(testtools.TestCase):

    def setUp(self):
        super(AZManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = az.AvailabilityZoneManager(self.api)

    def test_availability_zones_list(self):
        zones = self.mgr.list()
        expect = [
            ('GET', '/v1/availability_zones', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(zones, matchers.HasLength(2))



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import re
import sys
from unittest import mock

import fixtures
from keystoneauth1 import fixture
from testtools import matchers

from zunclient import api_versions
from zunclient import exceptions
import zunclient.shell
from zunclient.tests.unit import utils

FAKE_ENV = {'OS_USERNAME': 'username',
            'OS_PASSWORD': 'password',
            'OS_PROJECT_NAME': 'project_name',
            'OS_AUTH_URL': 'http://no.where/v2.0'}

FAKE_ENV2 = {'OS_USER_ID': 'user_id',
             'OS_PASSWORD': 'password',
             'OS_PROJECT_ID': 'project_id',
             'OS_AUTH_URL': 'http://no.where/v2.0'}

FAKE_ENV3 = {'OS_USERNAME': 'username',
             'OS_PASSWORD': 'password',
             'OS_PROJECT_ID': 'project_id',
             'OS_AUTH_URL': 'http://no.where/v2.0'}

FAKE_ENV4 = {'OS_USERNAME': 'username',
             'OS_PASSWORD': 'password',
             'OS_PROJECT_ID': 'project_id',
             'OS_USER_DOMAIN_NAME': 'Default',
             'OS_PROJECT_DOMAIN_NAME': 'Default',
             'OS_AUTH_URL': 'http://no.where/v3'}


def _create_ver_list(versions):
    return {'versions': {'values': versions}}


class ParserTest(utils.TestCase):

    def setUp(self):
        super(ParserTest, self).setUp()
        self.parser = zunclient.shell.ZunClientArgumentParser()

    def test_ambiguous_option(self):
        self.parser.add_argument('--tic')
        self.parser.add_argument('--tac')
        try:
            self.parser.parse_args(['--t'])
        except SystemExit as err:
            self.assertEqual(2, err.code)
        else:
            self.fail('SystemExit not raised')


class ShellTest(utils.TestCase):
    AUTH_URL = utils.FAKE_ENV['OS_AUTH_URL']

    _msg_no_tenant_project = ("You must provide a project name or project id"
                              " via --os-project-name, --os-project-id,"
                              " env[OS_PROJECT_NAME] or env[OS_PROJECT_ID]")

    def setUp(self):
        super(ShellTest, self).setUp()
        self.nc_util = mock.patch(
            'zunclient.common.cliutils.isunauthenticated').start()
        self.nc_util.return_value = False
        self.discover_version = mock.patch(
            'zunclient.api_versions.discover_version').start()
        self.discover_version.return_value = api_versions.APIVersion('1.1')

    def test_help_unknown_command(self):
        self.assertRaises(exceptions.CommandError, self.shell, 'help foofoo')

    def test_help(self):
        required = [
            r'.*?^usage: ',
            r'.*?^\s+stop\s+Stop specified container.',
            r'.*?^See "zun help COMMAND" for help on a specific command',
        ]
        stdout, stderr = self.shell('help')
        for r in required:
            self.assertThat((stdout + stderr),
                            matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))

    def test_help_on_subcommand(self):
        if sys.version_info >= (3, 10):
            options_name = 'Options'
        else:
            options_name = 'Optional arguments'
        required = [
            r'.*?^usage: zun create',
            r'.*?^Create a container.',
            r'.*?^' + options_name + ':',
        ]
        stdout, stderr = self.shell('help create')
        for r in required:
            self.assertThat((stdout + stderr),
                            matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))

    def test_help_no_options(self):
        required = [
            r'.*?^usage: ',
            r'.*?^\s+stop\s+Stop specified container.',
            r'.*?^See "zun help COMMAND" for help on a specific command',
        ]
        stdout, stderr = self.shell('')
        for r in required:
            self.assertThat((stdout + stderr),
                            matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))

    def test_bash_completion(self):
        stdout, stderr = self.shell('bash-completion')
        # just check we have some output
        required = [
            r'.*--format',
            r'.*help',
            r'.*show',
            r'.*--name']
        for r in required:
            self.assertThat((stdout + stderr),
                            matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))

    def test_no_username(self):
        required = ('You must provide a username via either'
                    ' --os-username or env[OS_USERNAME]')
        self.make_env(exclude='OS_USERNAME')
        try:
            self.shell('service-list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_user_id(self):
        required = ('You must provide a username via'
                    ' either --os-username or env[OS_USERNAME]')
        self.make_env(exclude='OS_USER_ID', fake_env=FAKE_ENV2)
        try:
            self.shell('service-list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_project_name(self):
        required = self._msg_no_tenant_project
        self.make_env(exclude='OS_PROJECT_NAME')
        try:
            self.shell('service-list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_project_id(self):
        required = self._msg_no_tenant_project
        self.make_env(exclude='OS_PROJECT_ID', fake_env=FAKE_ENV3)
        try:
            self.shell('service-list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_auth_url(self):
        required = ('You must provide an auth url'
                    ' via either --os-auth-url or env[OS_AUTH_URL] or'
                    ' specify an auth_system which defines a default url'
                    ' with --os-auth-system or env[OS_AUTH_SYSTEM]',)
        self.make_env(exclude='OS_AUTH_URL')
        try:
            self.shell('service-list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args)
        else:
            self.fail('CommandError not raised')

    @mock.patch('zunclient.client.Client')
    def test_service_type(self, mock_client):
        self.make_env()
        self.shell('service-list')
        _, client_kwargs = mock_client.call_args_list[0]
        self.assertEqual('container', client_kwargs['service_type'])

    @mock.patch('zunclient.client.Client')
    def test_container_format_env(self, mock_client):
        self.make_env()
        self.shell('create --environment key=value test')
        _, create_args = mock_client.return_value.containers.create.call_args
        self.assertEqual({'key': 'value'}, create_args['environment'])

    @mock.patch('zunclient.client.Client')
    def test_insecure(self, mock_client):
        self.make_env()
        self.shell('--insecure service-list')
        _, session_kwargs = mock_client.call_args_list[0]
        self.assertEqual(True, session_kwargs['insecure'])

    @mock.patch('sys.stdin', side_effect=mock.MagicMock)
    @mock.patch('getpass.getpass', side_effect=EOFError)
    def test_no_password(self, mock_getpass, mock_stdin):
        required = ('Expecting a password provided'
                    ' via either --os-password, env[OS_PASSWORD],'
                    ' or prompted response',)
        self.make_env(exclude='OS_PASSWORD')
        try:
            self.shell('service-list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args)
        else:
            self.fail('CommandError not raised')

    @mock.patch('sys.argv', ['zun'])
    @mock.patch('sys.stdout', io.StringIO())
    @mock.patch('sys.stderr', io.StringIO())
    def test_main_noargs(self):
        # Ensure that main works with no command-line arguments
        try:
            zunclient.shell.main()
        except SystemExit:
            self.fail('Unexpected SystemExit')

        # We expect the normal usage as a result
        self.assertIn('Command-line interface to the OpenStack Zun API',
                      sys.stdout.getvalue())

    @mock.patch('zunclient.client.Client')
    def _test_main_region(self, command, expected_region_name, mock_client):
        self.shell(command)
        mock_client.assert_called_once_with(
            username='username', password='password',
            interface='publicURL', project_id=None,
            project_name='project_name', auth_url=self.AUTH_URL,
            service_type='container', region_name=expected_region_name,
            project_domain_id='', project_domain_name='',
            user_domain_id='', user_domain_name='', profile=None,
            endpoint_override=None, insecure=False, cacert=None,
            cert=None, key=None,
            version=api_versions.APIVersion('1.29'))

    def test_main_option_region(self):
        self.make_env()
        self._test_main_region(
            '--zun-api-version 1.29 '
            '--os-region-name=myregion service-list', 'myregion')

    def test_main_env_region(self):
        fake_env = dict(utils.FAKE_ENV, OS_REGION_NAME='myregion')
        self.make_env(fake_env=fake_env)
        self._test_main_region(
            '--zun-api-version 1.29 '
            'service-list', 'myregion')

    def test_main_no_region(self):
        self.make_env()
        self._test_main_region(
            '--zun-api-version 1.29 '
            'service-list', None)

    @mock.patch('zunclient.client.Client')
    def test_main_endpoint_public(self, mock_client):
        self.make_env()
        self.shell(
            '--zun-api-version 1.29 '
            '--endpoint-type publicURL service-list')
        mock_client.assert_called_once_with(
            username='username', password='password',
            interface='publicURL', project_id=None,
            project_name='project_name', auth_url=self.AUTH_URL,
            service_type='container', region_name=None,
            project_domain_id='', project_domain_name='',
            user_domain_id='', user_domain_name='', profile=None,
            endpoint_override=None, insecure=False, cacert=None,
            cert=None, key=None,
            version=api_versions.APIVersion('1.29'))

    @mock.patch('zunclient.client.Client')
    def test_main_endpoint_internal(self, mock_client):
        self.make_env()
        self.shell(
            '--zun-api-version 1.29 '
            '--endpoint-type internalURL service-list')
        mock_client.assert_called_once_with(
            username='username', password='password',
            interface='internalURL', project_id=None,
            project_name='project_name', auth_url=self.AUTH_URL,
            service_type='container', region_name=None,
            project_domain_id='', project_domain_name='',
            user_domain_id='', user_domain_name='', profile=None,
            endpoint_override=None, insecure=False, cacert=None,
            cert=None, key=None,
            version=api_versions.APIVersion('1.29'))


class ShellTestKeystoneV3(ShellTest):
    AUTH_URL = 'http://no.where/v3'

    def make_env(self, exclude=None, fake_env=FAKE_ENV):
        if 'OS_AUTH_URL' in fake_env:
            fake_env.update({'OS_AUTH_URL': self.AUTH_URL})
        env = dict((k, v) for k, v in fake_env.items() if k != exclude)
        self.useFixture(fixtures.MonkeyPatch('os.environ', env))

    def register_keystone_discovery_fixture(self, mreq):
        v3_url = "http://no.where/v3"
        v3_version = fixture.V3Discovery(v3_url)
        mreq.register_uri(
            'GET', v3_url, json=_create_ver_list([v3_version]),
            status_code=200)

    @mock.patch('zunclient.client.Client')
    def test_main_endpoint_public(self, mock_client):
        self.make_env(fake_env=FAKE_ENV4)
        self.shell(
            '--zun-api-version 1.29 '
            '--endpoint-type publicURL service-list')
        mock_client.assert_called_once_with(
            username='username', password='password',
            interface='publicURL', project_id='project_id',
            project_name=None, auth_url=self.AUTH_URL,
            service_type='container', region_name=None,
            project_domain_id='', project_domain_name='Default',
            user_domain_id='', user_domain_name='Default',
            endpoint_override=None, insecure=False, profile=None,
            cacert=None, cert=None, key=None,
            version=api_versions.APIVersion('1.29'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain a
#    copy of the License at
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from zunclient.tests.unit import utils
from zunclient.v1 import quotas


DEFAULT_QUOTAS = {
    'containers': '40',
    'memory': '51200',
    'cpu': '20',
    'disk': '100'
}

MODIFIED_QUOTAS = {
    'containers': '50',
    'memory': '51200',
    'cpu': '20',
    'disk': '100'
}

MODIFIED_USAGE_QUOTAS = {
    'containers': {
        'limit': '50',
        'in_use': '30'
    },
    'memory': {},
    'cpu': {},
    'disk': {}
}

fake_responses = {
    '/v1/quotas/test_project_id':
    {
        'GET': (
            {},
            MODIFIED_QUOTAS
        ),
        'PUT': (
            {},
            MODIFIED_QUOTAS
        ),
        'DELETE': (
            {},
            None
        )
    },
    '/v1/quotas/test_project_id/defaults':
    {
        'GET': (
            {},
            DEFAULT_QUOTAS
        )
    },
    '/v1/quotas/test_project_id?usages=True':
    {
        'GET': (
            {},
            MODIFIED_USAGE_QUOTAS
        )
    }
}


class QuotaManagerTest(testtools.TestCase):

    def setUp(self):
        super(QuotaManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = quotas.QuotaManager(self.api)

    def test_quotas_get_defaults(self):
        quotas = self.mgr.defaults('test_project_id')
        expect = [
            ('GET', '/v1/quotas/test_project_id/defaults', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(quotas.containers, DEFAULT_QUOTAS['containers'])
        self.assertEqual(quotas.memory, DEFAULT_QUOTAS['memory'])
        self.assertEqual(quotas.cpu, DEFAULT_QUOTAS['cpu'])
        self.assertEqual(quotas.disk, DEFAULT_QUOTAS['disk'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from testtools import matchers

from zunclient.tests.unit import utils
from zunclient.v1 import services


SERVICE1 = {'id': 123,
            'host': 'fake-host1',
            'binary': 'fake-bin1',
            'state': 'up',
            'availability_zone': 'nova',
            }
SERVICE2 = {'id': 124,
            'host': 'fake-host2',
            'binary': 'fake-bin2',
            'state': 'down',
            'availability_zone': 'nova',
            }

fake_responses = {
    '/v1/services':
    {
        'GET': (
            {},
            {'services': [SERVICE1, SERVICE2]},
        ),
    },
    '/v1/services/?limit=2':
    {
        'GET': (
            {},
            {'services': [SERVICE1, SERVICE2]},
        ),
    },
    '/v1/services/?marker=%s' % SERVICE2['id']:
    {
        'GET': (
            {},
            {'services': [SERVICE1, SERVICE2]},
        ),
    },
    '/v1/services/?limit=2&marker=%s' % SERVICE2['id']:
    {
        'GET': (
            {},
            {'services': [SERVICE2, SERVICE1]},
        ),
    },
    '/v1/services/?sort_dir=asc':
    {
        'GET': (
            {},
            {'services': [SERVICE1, SERVICE2]},
        ),
    },
    '/v1/services/?sort_key=id':
    {
        'GET': (
            {},
            {'services': [SERVICE1, SERVICE2]},
        ),
    },
    '/v1/services/?sort_key=id&sort_dir=desc':
    {
        'GET': (
            {},
            {'services': [SERVICE2, SERVICE1]},
        ),
    },
}


class ServiceManagerTest(testtools.TestCase):

    def setUp(self):
        super(ServiceManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = services.ServiceManager(self.api)

    def test_service_list(self):
        services = self.mgr.list()
        expect = [
            ('GET', '/v1/services', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(services, matchers.HasLength(2))

    def _test_service_list_with_filters(
            self, limit=None, marker=None,
            sort_key=None, sort_dir=None,
            expect=[]):
        services_filter = self.mgr.list(limit=limit, marker=marker,
                                        sort_key=sort_key,
                                        sort_dir=sort_dir)
        self.assertEqual(expect, self.api.calls)
        self.assertThat(services_filter, matchers.HasLength(2))

    def test_service_list_with_limit(self):
        expect = [
            ('GET', '/v1/services/?limit=2', {}, None),
        ]
        self._test_service_list_with_filters(
            limit=2,
            expect=expect)

    def test_service_list_with_marker(self):
        expect = [
            ('GET', '/v1/services/?marker=%s' % SERVICE2['id'],
             {}, None),
        ]
        self._test_service_list_with_filters(
            marker=SERVICE2['id'],
            expect=expect)

    def test_service_list_with_marker_limit(self):
        expect = [
            ('GET', '/v1/services/?limit=2&marker=%s' % SERVICE2['id'],
             {}, None),
        ]
        self._test_service_list_with_filters(
            limit=2, marker=SERVICE2['id'],
            expect=expect)

    def test_service_list_with_sort_dir(self):
        expect = [
            ('GET', '/v1/services/?sort_dir=asc',
             {}, None),
        ]
        self._test_service_list_with_filters(
            sort_dir='asc',
            expect=expect)

    def test_service_list_with_sort_key(self):
        expect = [
            ('GET', '/v1/services/?sort_key=id',
             {}, None),
        ]
        self._test_service_list_with_filters(
            sort_key='id',
            expect=expect)

    def test_service_list_with_sort_key_dir(self):
        expect = [
            ('GET', '/v1/services/?sort_key=id&sort_dir=desc',
             {}, None),
        ]
        self._test_service_list_with_filters(
            sort_key='id', sort_dir='desc',
            expect=expect)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from testtools import matchers

from zunclient.tests.unit import utils
from zunclient.v1 import versions


VERSION1 = {'status': 'CURRENT',
            'min_version': '1.1',
            'max_version': '1.12',
            'id': 'v1',
            }

fake_responses = {
    'https://example.com':
    {
        'GET': (
            {},
            {'versions': [VERSION1]},
        ),
    },
}


class VersionManagerTest(testtools.TestCase):

    def test_version_list(self):
        self._test_version_list('https://example.com')
        self._test_version_list('https://example.com/v1')
        self._test_version_list('https://example.com/v1/')

    def _test_version_list(self, endpoint):
        api = utils.FakeAPI(fake_responses, endpoint=endpoint)
        mgr = versions.VersionManager(api)
        api_versions = mgr.list()
        expect = [
            ('GET', 'https://example.com', {}, None),
        ]
        self.assertEqual(expect, api.calls)
        self.assertThat(api_versions, matchers.HasLength(1))



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from zunclient.common import cliutils as utils
from zunclient.common import utils as zun_utils


def do_service_list(cs, args):
    """Print a list of zun services."""
    services = cs.services.list()
    columns = ('Id', 'Host', 'Binary', 'State', 'Disabled',
               'Disabled Reason', 'Updated At',
               'Availability Zone')
    utils.print_list(services, columns,
                     {'versions': zun_utils.print_list_field('versions')})


@utils.arg('host',
           metavar='<host>',
           help='Name of host.')
@utils.arg('binary',
           metavar='<binary>',
           help='Name of the binary to delete.')
def do_service_delete(cs, args):
    """Delete the Zun binaries/services."""
    try:
        cs.services.delete(args.host, args.binary)
        print("Request to delete binary %s on host %s has been accepted." %
              (args.binary, args.host))
    except Exception as e:
        print("Delete for binary %(binary)s on host %(host)s failed: %(e)s" %
              {'binary': args.binary, 'host': args.host, 'e': e})


@utils.arg('host', metavar='<host>', help='Name of host.')
@utils.arg('binary', metavar='<binary>', help='Service binary.')
def do_service_enable(cs, args):
    """Enable the Zun service."""
    res = cs.services.enable(args.host, args.binary)
    utils.print_dict(res[1]['service'])


@utils.arg('host', metavar='<host>', help='Name of host.')
@utils.arg('binary', metavar='<binary>', help='Service binary.')
@utils.arg(
    '--reason',
    metavar='<reason>',
    help='Reason for disabling service.')
def do_service_disable(cs, args):
    """Disable the Zun service."""
    res = cs.services.disable(args.host, args.binary, args.reason)
    utils.print_dict(res[1]['service'])


@utils.arg('host', metavar='<host>', help='Name of host.')
@utils.arg('binary', metavar='<binary>', help='Service binary.')
@utils.arg(
    '--unset',
    dest='force_down',
    help="Unset the force state down of service.",
    action='store_false',
    default=True)
def do_service_force_down(cs, args):
    """Force Zun service to down or unset the force state."""
    res = cs.services.force_down(args.host, args.binary, args.force_down)
    utils.print_dict(res[1]['service'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zunclient.common import cliutils as utils


@utils.arg(
    '--containers',
    metavar='<containers>',
    type=int,
    help='The number of containers allowed per project')
@utils.arg(
    '--cpu',
    metavar='<cpu>',
    type=int,
    help='The number of container cores or vCPUs allowed per project')
@utils.arg(
    '--memory',
    metavar='<memory>',
    type=int,
    help='The number of megabytes of container RAM allowed per project')
@utils.arg(
    '--disk',
    metavar='<disk>',
    type=int,
    help='The number of gigabytes of container Disk allowed per project')
@utils.arg(
    'quota_class_name',
    metavar='<quota_class_name>',
    help='The name of quota class')
def do_quota_class_update(cs, args):
    """Print an updated quotas for a quota class"""
    utils.print_dict(cs.quota_classes.update(
        args.quota_class_name,
        containers=args.containers,
        memory=args.memory,
        cpu=args.cpu,
        disk=args.disk)._info)


@utils.arg(
    'quota_class_name',
    metavar='<quota_class_name>',
    help='The name of quota class')
def do_quota_class_get(cs, args):
    """Print a quotas for a quota class"""
    utils.print_dict(cs.quota_classes.get(args.quota_class_name)._info)



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from contextlib import closing
import io
import os
import tarfile
import time
import yaml

from oslo_serialization import jsonutils

from zunclient.common import cliutils as utils
from zunclient.common import utils as zun_utils
from zunclient.common.websocketclient import exceptions
from zunclient.common.websocketclient import websocketclient
from zunclient import exceptions as exc


RENAME_DEPRECATION_MESSAGE = (
    'WARNING: Rename container command deprecated and will be removed '
    'in a future release.\nUse Update container command to avoid '
    'seeing this message.')


SG_DEPRECATION_MESSAGE = (
    'WARNING: Security group related commands deprecated and will be removed '
    'in a future release.\nUse Neutron commands to manage security groups '
    'instead.')


def _show_container(container):
    zun_utils.format_container_addresses(container)
    utils.print_dict(container._info)


@utils.exclusive_arg(
    'restart_auto_remove',
    '--auto-remove',
    required=False, action='store_true',
    help='Automatically remove the container when it exits')
@utils.exclusive_arg(
    'restart_auto_remove',
    '--restart',
    required=False, metavar='<restart>',
    help='Restart policy to apply when a container exits'
         '(no, on-failure[:max-retry], always, unless-stopped)')
@utils.exclusive_arg(
    'secgroup_expose_port',
    '--security-group',
    metavar='<security-group>',
    action='append', default=[],
    help='The name of security group for the container. '
         'May be used multiple times.')
@utils.exclusive_arg(
    'secgroup_expose_port',
    '-p', '--expose-port',
    action='append',
    default=[],
    metavar='<port>',
    help='Expose container port(s) to outside (format: <port>[/<protocol>])')
@utils.arg('-n', '--name',
           metavar='<name>',
           help='name of the container')
@utils.arg('--cpu',
           metavar='<cpu>',
           help='The number of virtual cpus.')
@utils.arg('-m', '--memory',
           metavar='<memory>',
           help='The container memory size in MiB')
@utils.arg('-e', '--environment',
           metavar='<KEY=VALUE>',
           action='append', default=[],
           help='The environment variables')
@utils.arg('--workdir',
           metavar='<workdir>',
           help='The working directory for commands to run in')
@utils.arg('--label',
           metavar='<KEY=VALUE>',
           action='append', default=[],
           help='Adds a map of labels to a container. '
                'May be used multiple times.')
@utils.arg('--image-pull-policy',
           dest='image_pull_policy',
           metavar='<policy>',
           choices=['never', 'always', 'ifnotpresent'],
           help='The policy which determines if the image should '
                'be pulled prior to starting the container. '
                'It can have following values: '
                '"ifnotpresent": only pull the image if it does not '
                'already exist on the node. '
                '"always": Always pull the image from repository.'
                '"never": never pull the image')
@utils.arg('image', metavar='<image>', help='name or ID or repo of the image '
           '(e.g. cirros:latest)')
@utils.arg('-i', '--interactive',
           dest='interactive',
           action='store_true',
           default=False,
           help='Keep STDIN open even if not attached, allocate a pseudo-TTY')
@utils.arg('--image-driver',
           metavar='<image_driver>',
           help='The image driver to use to pull container image. '
                'It can have following values: '
                '"docker": pull the image from Docker Hub. '
                '"glance": pull the image from Glance. ')
@utils.arg('command',
           metavar='<command>',
           nargs=argparse.REMAINDER,
           help='Send command to the container')
@utils.arg('--hint',
           action='append',
           default=[],
           metavar='<key=value>',
           help='The key-value pair(s) for scheduler to select host. '
                'The format of this parameter is "key=value[,key=value]". '
                'May be used multiple times.')
@utils.arg('--net',
           action='append',
           default=[],
           metavar='<network=network, port=port-uuid,'
                   'v4-fixed-ip=ip-addr,v6-fixed-ip=ip-addr>',
           help='Create network enpoints for the container. '
                'network: attach container to the specified neturon networks. '
                'port: attach container to the neutron port with this UUID. '
                'v4-fixed-ip: IPv4 fixed address for container. '
                'v6-fixed-ip: IPv6 fixed address for container.')
@utils.arg('--mount',
           action='append',
           default=[],
           metavar='<mount>',
           help='A dictionary to configure volumes mounted inside the '
                'container.')
@utils.arg('--runtime',
           metavar='<runtime>',
           help='The runtime to use for this container. '
                'It can have value "runc" or any other custom runtime.')
@utils.arg('--hostname',
           metavar='<hostname>',
           default=None,
           help='Container host name')
@utils.arg('--disk',
           metavar='<disk>',
           type=int,
           default=None,
           help='The disk size in GiB for per container.')
@utils.arg('--availability-zone',
           metavar='<availability_zone>',
           default=None,
           help='The availability zone of the container.')
@utils.arg('--auto-heal',
           action='store_true',
           default=False,
           help='The flag of healing non-existent container in docker.')
@utils.arg('--privileged',
           dest='privileged',
           action='store_true',
           default=False,
           help='Give extended privileges to this container')
@utils.arg('--healthcheck',
           action='append',
           default=[],
           metavar='<cmd=command,interval=time,retries=integer,timeout=time>',
           help='Specify a test cmd to perform to check that the container'
                'is healthy. '
                'cmd: Command to run to check health. '
                'interval: Time between running the check (s|m|h)'
                '          (default 0s). '
                'retries: Consecutive failures needed to report unhealthy. '
                'timeout: Maximum time to allow one check to run (s|m|h)'
                '         (default 0s).')
@utils.arg('--registry',
           metavar='<registry>',
           help='The container image registry ID or name')
@utils.arg('--host',
           metavar='<host>',
           help='Requested host to create containers. Admin only by default.'
                '(Supported by API versions 1.39 or above)')
@utils.arg('--entrypoint',
           metavar='<entrypoint>',
           help='The entrypoint which overwrites the default ENTRYPOINT '
                'of the image. (Supported by API versions 1.40 or above)')
def do_create(cs, args):
    """Create a container."""
    opts = {}
    opts['name'] = args.name
    opts['image'] = args.image
    opts['memory'] = args.memory
    opts['cpu'] = args.cpu
    opts['environment'] = zun_utils.format_args(args.environment)
    opts['auto_remove'] = args.auto_remove
    opts['workdir'] = args.workdir
    opts['labels'] = zun_utils.format_args(args.label)
    opts['image_pull_policy'] = args.image_pull_policy
    opts['image_driver'] = args.image_driver
    opts['hints'] = zun_utils.format_args(args.hint)
    opts['nets'] = zun_utils.parse_nets(args.net)
    opts['mounts'] = zun_utils.parse_mounts(args.mount)
    opts['runtime'] = args.runtime
    opts['hostname'] = args.hostname
    opts['disk'] = args.disk
    opts['availability_zone'] = args.availability_zone
    opts['command'] = args.command
    opts['registry'] = args.registry
    opts['host'] = args.host
    if args.entrypoint:
        opts['entrypoint'] = zun_utils.parse_entrypoint(args.entrypoint)
    if args.healthcheck:
        opts['healthcheck'] = zun_utils.parse_health(args.healthcheck)

    if args.auto_heal:
        opts['auto_heal'] = args.auto_heal
    if args.security_group:
        opts['security_groups'] = args.security_group
    if args.expose_port:
        opts['exposed_ports'] = zun_utils.parse_exposed_ports(args.expose_port)
    if args.restart:
        opts['restart_policy'] = zun_utils.check_restart_policy(args.restart)
    if args.interactive:
        opts['interactive'] = True
    if args.privileged:
        opts['privileged'] = True
    opts = zun_utils.remove_null_parms(**opts)
    _show_container(cs.containers.create(**opts))


@utils.arg('--all-projects',
           action="store_true",
           default=False,
           help='List containers in all projects')
@utils.arg('--marker',
           metavar='<marker>',
           default=None,
           help='The last container UUID of the previous page; '
                'displays list of containers after "marker".')
@utils.arg('--limit',
           metavar='<limit>',
           type=int,
           help='Maximum number of containers to return')
@utils.arg('--sort-key',
           metavar='<sort-key>',
           help='Column to sort results by')
@utils.arg('--sort-dir',
           metavar='<sort-dir>',
           choices=['desc', 'asc'],
           help='Direction to sort. "asc" or "desc".')
@utils.arg('--name',
           metavar='<name>',
           help='List containers according to their name.')
@utils.arg('--image',
           metavar='<image>',
           help='List containers according to their image.')
@utils.arg('--project-id',
           metavar='<project-id>',
           help='List containers according to their Project_id')
@utils.arg('--user-id',
           metavar='<user-id>',
           help='List containers according to their user_id')
@utils.arg('--task-state',
           metavar='<task-state>',
           help='List containers according to their task-state')
@utils.arg('--status',
           metavar='<status>',
           help='List containers according to their status')
@utils.arg('--memory',
           metavar='<memory>',
           help='List containers according to their memory size in MiB')
@utils.arg('--host',
           metavar='<host>',
           help='List containers according to their hostname')
@utils.arg('--auto-remove',
           metavar='<auto-remove>',
           help='List containers according to whether they are '
                'auto-removed on exiting')
def do_list(cs, args):
    """Print a list of available containers."""
    opts = {}
    opts['all_projects'] = args.all_projects
    opts['marker'] = args.marker
    opts['limit'] = args.limit
    opts['sort_key'] = args.sort_key
    opts['sort_dir'] = args.sort_dir
    opts['image'] = args.image
    opts['name'] = args.name
    opts['project_id'] = args.project_id
    opts['user_id'] = args.user_id
    opts['host'] = args.host
    opts['task_state'] = args.task_state
    opts['memory'] = args.memory
    opts['auto_remove'] = args.auto_remove
    opts['status'] = args.status
    opts = zun_utils.remove_null_parms(**opts)
    containers = cs.containers.list(**opts)
    zun_utils.list_containers(containers)


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID or name of the (container)s to delete.')
@utils.arg('-f', '--force',
           action='store_true',
           help='Force delete the container.')
@utils.arg('-s', '--stop',
           action='store_true',
           help='Stop the running container first before delete.')
@utils.arg('--all-projects',
           action="store_true",
           default=False,
           help='Delete container(s) in all projects by name.')
def do_delete(cs, args):
    """Delete specified containers."""
    for container in args.containers:
        opts = {}
        opts['id'] = container
        opts['force'] = args.force
        opts['stop'] = args.stop
        opts['all_projects'] = args.all_projects
        opts = zun_utils.remove_null_parms(**opts)
        try:
            cs.containers.delete(**opts)
            print("Request to delete container %s has been accepted." %
                  container)
        except Exception as e:
            print("Delete for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to show.')
@utils.arg('-f', '--format',
           metavar='<format>',
           action='store',
           choices=['json', 'yaml', 'table'],
           default='table',
           help='Print representation of the container.'
                'The choices of the output format is json,table,yaml.'
                'Defaults to table.')
@utils.arg('--all-projects',
           action="store_true",
           default=False,
           help='Show container(s) in all projects by name.')
def do_show(cs, args):
    """Show details of a container."""
    opts = {}
    opts['id'] = args.container
    opts['all_projects'] = args.all_projects
    opts = zun_utils.remove_null_parms(**opts)
    container = cs.containers.get(**opts)
    if args.format == 'json':
        print(jsonutils.dumps(container._info, indent=4, sort_keys=True))
    elif args.format == 'yaml':
        print(yaml.safe_dump(container._info, default_flow_style=False))
    elif args.format == 'table':
        _show_container(container)


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID of the (container)s to rebuild.')
@utils.arg('--image',
           metavar='<image>',
           help='The image for specified container to update.')
@utils.arg('--image-driver',
           metavar='<image_driver>',
           help='The image driver to use to pull container image. '
                'It can have following values: '
                '"docker": pull the image from Docker Hub. '
                '"glance": pull the image from Glance. '
                'The default value is source container\'s image driver ')
def do_rebuild(cs, args):
    """Rebuild specified containers."""
    for container in args.containers:
        opts = {}
        opts['id'] = container
        if args.image:
            opts['image'] = args.image
        if args.image_driver:
            opts['image_driver'] = args.image_driver
        try:
            cs.containers.rebuild(**opts)
            print("Request to rebuild container %s has been accepted." %
                  container)
        except Exception as e:
            print("Rebuild for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID or name of the (container)s to restart.')
@utils.arg('-t', '--timeout',
           metavar='<timeout>',
           default=10,
           help='Seconds to wait for stop before restarting (container)s')
def do_restart(cs, args):
    """Restart specified containers."""
    for container in args.containers:
        try:
            cs.containers.restart(container, args.timeout)
            print("Request to restart container %s has been accepted." %
                  container)
        except Exception as e:
            print("Restart for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID or name of the (container)s to stop.')
@utils.arg('-t', '--timeout',
           metavar='<timeout>',
           default=10,
           help='Seconds to wait for stop before killing (container)s')
def do_stop(cs, args):
    """Stop specified containers."""
    for container in args.containers:
        try:
            cs.containers.stop(container, args.timeout)
            print("Request to stop container %s has been accepted." %
                  container)
        except Exception as e:
            print("Stop for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID of the (container)s to start.')
def do_start(cs, args):
    """Start specified containers."""
    for container in args.containers:
        try:
            cs.containers.start(container)
            print("Request to start container %s has been accepted." %
                  container)
        except Exception as e:
            print("Start for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID or name of the (container)s to pause.')
def do_pause(cs, args):
    """Pause specified containers."""
    for container in args.containers:
        try:
            cs.containers.pause(container)
            print("Request to pause container %s has been accepted." %
                  container)
        except Exception as e:
            print("Pause for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID or name of the (container)s to unpause.')
def do_unpause(cs, args):
    """Unpause specified containers."""
    for container in args.containers:
        try:
            cs.containers.unpause(container)
            print("Request to unpause container %s has been accepted." %
                  container)
        except Exception as e:
            print("Unpause for container %(container)s failed: %(e)s" %
                  {'container': container, 'e': e})


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to get logs for.')
@utils.arg('--stdout',
           action='store_true',
           help='Only stdout logs of container.')
@utils.arg('--stderr',
           action='store_true',
           help='Only stderr logs of container.')
@utils.arg('--since',
           metavar='<since>',
           default=None,
           help='Show logs since a given datetime or integer '
                'epoch (in seconds).')
@utils.arg('-t', '--timestamps',
           dest='timestamps',
           action='store_true',
           default=False,
           help='Show timestamps.')
@utils.arg('--tail',
           metavar='<tail>',
           default='all',
           help='Number of lines to show from the end of the logs.')
def do_logs(cs, args):
    """Get logs of a container."""
    opts = {}
    opts['id'] = args.container
    opts['stdout'] = args.stdout
    opts['stderr'] = args.stderr
    opts['since'] = args.since
    opts['timestamps'] = args.timestamps
    opts['tail'] = args.tail
    opts = zun_utils.remove_null_parms(**opts)
    logs = cs.containers.logs(**opts)
    print(logs)


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to execute command in.')
@utils.arg('command',
           metavar='<command>',
           nargs=argparse.REMAINDER,
           help='The command to execute in a container')
@utils.arg('-i', '--interactive',
           dest='interactive',
           action='store_true',
           default=False,
           help='Keep STDIN open and allocate a pseudo-TTY for interactive')
def do_exec(cs, args):
    """Execute command in a running container."""
    opts = {}
    opts['command'] = zun_utils.parse_command(args.command)
    if args.interactive:
        opts['interactive'] = True
        opts['run'] = False
    response = cs.containers.execute(args.container, **opts)
    if args.interactive:
        exec_id = response['exec_id']
        url = response['proxy_url']
        websocketclient.do_exec(cs, url, args.container, exec_id, "~", 0.5)
    else:
        output = response['output']
        exit_code = response['exit_code']
        print(output)
        return exit_code


@utils.arg('containers',
           metavar='<container>',
           nargs='+',
           help='ID or name of the (container)s to kill signal to.')
@utils.arg('-s', '--signal',
           metavar='<signal>',
           default=None,
           help='The signal to kill')
def do_kill(cs, args):
    """Kill one or more running container(s)."""
    for container in args.containers:
        opts = {}
        opts['id'] = container
        opts['signal'] = args.signal
        opts = zun_utils.remove_null_parms(**opts)
        try:
            cs.containers.kill(**opts)
            print(
                "Request to kill signal to container %s has been accepted." %
                container)
        except Exception as e:
            print(
                "kill signal for container %(container)s failed: %(e)s" %
                {'container': container, 'e': e})


@utils.exclusive_arg(
    'restart_auto_remove',
    '--auto-remove',
    required=False, action='store_true',
    help='Automatically remove the container when it exits')
@utils.exclusive_arg(
    'restart_auto_remove',
    '--restart',
    required=False, metavar='<restart>',
    help='Restart policy to apply when a container exits'
         '(no, on-failure[:max-retry], always, unless-stopped)')
@utils.exclusive_arg(
    'secgroup_expose_port',
    '--security-group',
    metavar='<security-group>',
    action='append', default=[],
    help='The name of security group for the container. '
         'May be used multiple times.')
@utils.exclusive_arg(
    'secgroup_expose_port',
    '-p', '--expose-port',
    action='append',
    default=[],
    metavar='<port>',
    help='Expose container port(s) to outside (format: <port>[/<protocol>])')
@utils.arg('-n', '--name',
           metavar='<name>',
           help='name of the container')
@utils.arg('--cpu',
           metavar='<cpu>',
           help='The number of virtual cpus.')
@utils.arg('-m', '--memory',
           metavar='<memory>',
           help='The container memory size in MiB')
@utils.arg('-e', '--environment',
           metavar='<KEY=VALUE>',
           action='append', default=[],
           help='The environment variables')
@utils.arg('--workdir',
           metavar='<workdir>',
           help='The working directory for commands to run in')
@utils.arg('--label',
           metavar='<KEY=VALUE>',
           action='append', default=[],
           help='Adds a map of labels to a container. '
                'May be used multiple times.')
@utils.arg('--image-pull-policy',
           dest='image_pull_policy',
           metavar='<policy>',
           choices=['never', 'always', 'ifnotpresent'],
           help='The policy which determines if the image should '
                'be pulled prior to starting the container. '
                'It can have following values: '
                '"ifnotpresent": only pull the image if it does not '
                'already exist on the node. '
                '"always": Always pull the image from repository.'
                '"never": never pull the image')
@utils.arg('image', metavar='<image>', help='name or ID of the image')
@utils.arg('-i', '--interactive',
           dest='interactive',
           action='store_true',
           default=False,
           help='Keep STDIN open even if not attached, allocate a pseudo-TTY')
@utils.arg('--image-driver',
           metavar='<image_driver>',
           help='The image driver to use to pull container image. '
                'It can have following values: '
                '"docker": pull the image from Docker Hub. '
                '"glance": pull the image from Glance. ')
@utils.arg('command',
           metavar='<command>',
           nargs=argparse.REMAINDER,
           help='Send command to the container')
@utils.arg('--hint',
           action='append',
           default=[],
           metavar='<key=value>',
           help='The key-value pair(s) for scheduler to select host. '
                'The format of this parameter is "key=value[,key=value]". '
                'May be used multiple times.')
@utils.arg('--net',
           action='append',
           default=[],
           metavar='<network=network, port=port-uuid,'
                   'v4-fixed-ip=ip-addr,v6-fixed-ip=ip-addr>',
           help='Create network enpoints for the container. '
                'network: attach container to the specified neutron networks. '
                'port: attach container to the neutron port with this UUID. '
                'v4-fixed-ip: IPv4 fixed address for container. '
                'v6-fixed-ip: IPv6 fixed address for container.')
@utils.arg('--mount',
           action='append',
           default=[],
           metavar='<mount>',
           help='A dictionary to configure volumes mounted inside the '
                'container.')
@utils.arg('--runtime',
           metavar='<runtime>',
           help='The runtime to use for this container. '
                'It can have value "runc" or any other custom runtime.')
@utils.arg('--hostname',
           metavar='<hostname>',
           default=None,
           help='Container hostname')
@utils.arg('--disk',
           metavar='<disk>',
           type=int,
           default=None,
           help='The disk size in GiB for per container.')
@utils.arg('--availability-zone',
           metavar='<availability_zone>',
           default=None,
           help='The availability zone of the container.')
@utils.arg('--auto-heal',
           action='store_true',
           default=False,
           help='The flag of healing non-existent container in docker.')
@utils.arg('--privileged',
           dest='privileged',
           action='store_true',
           default=False,
           help='Give extended privileges to this container')
@utils.arg('--healthcheck',
           action='append',
           default=[],
           metavar='<cmd=command,interval=time,retries=integer,timeout=time>',
           help='Specify a test cmd to perform to check that the container'
                'is healthy. '
                'cmd: Command to run to check health. '
                'interval: Time between running the check (s|m|h)'
                '          (default 0s). '
                'retries: Consecutive failures needed to report unhealthy. '
                'timeout: Maximum time to allow one check to run (s|m|h)'
                '         (default 0s).')
@utils.arg('--registry',
           metavar='<registry>',
           help='The container image registry ID or name')
@utils.arg('--host',
           metavar='<host>',
           help='Requested host to run containers. Admin only by default.'
                '(Supported by API versions 1.39 or above)')
@utils.arg('--entrypoint',
           metavar='<entrypoint>',
           help='The entrypoint which overwrites the default ENTRYPOINT '
                'of the image. (Supported by API versions 1.40 or above)')
def do_run(cs, args):
    """Run a command in a new container."""
    opts = {}
    opts['name'] = args.name
    opts['image'] = args.image
    opts['memory'] = args.memory
    opts['cpu'] = args.cpu
    opts['environment'] = zun_utils.format_args(args.environment)
    opts['workdir'] = args.workdir
    opts['auto_remove'] = args.auto_remove
    opts['labels'] = zun_utils.format_args(args.label)
    opts['image_pull_policy'] = args.image_pull_policy
    opts['image_driver'] = args.image_driver
    opts['hints'] = zun_utils.format_args(args.hint)
    opts['nets'] = zun_utils.parse_nets(args.net)
    opts['mounts'] = zun_utils.parse_mounts(args.mount)
    opts['runtime'] = args.runtime
    opts['hostname'] = args.hostname
    opts['disk'] = args.disk
    opts['availability_zone'] = args.availability_zone
    opts['command'] = args.command
    opts['registry'] = args.registry
    opts['host'] = args.host
    if args.entrypoint:
        opts['entrypoint'] = zun_utils.parse_entrypoint(args.entrypoint)
    if args.healthcheck:
        opts['healthcheck'] = zun_utils.parse_health(args.healthcheck)

    if args.auto_heal:
        opts['auto_heal'] = args.auto_heal
    if args.security_group:
        opts['security_groups'] = args.security_group
    if args.expose_port:
        opts['exposed_ports'] = zun_utils.parse_exposed_ports(args.expose_port)
    if args.restart:
        opts['restart_policy'] = zun_utils.check_restart_policy(args.restart)
    if args.interactive:
        opts['interactive'] = True
    if args.privileged:
        opts['privileged'] = True
    opts = zun_utils.remove_null_parms(**opts)
    container = cs.containers.run(**opts)
    _show_container(container)
    container_uuid = getattr(container, 'uuid', None)
    if args.interactive:
        ready_for_attach = False
        while True:
            container = cs.containers.get(container_uuid)
            if zun_utils.check_container_status(container, 'Running'):
                ready_for_attach = True
                break
            if zun_utils.check_container_status(container, 'Error'):
                raise exceptions.ContainerStateError(container_uuid)
            print("Waiting for container start")
            time.sleep(1)
        if ready_for_attach is True:
            response = cs.containers.attach(container_uuid)
            websocketclient.do_attach(cs, response, container_uuid, "~", 0.5)
        else:
            raise exceptions.InvalidWebSocketLink(container_uuid)


@utils.arg('container',
           metavar='<container>',
           help="ID or name of the container to update.")
@utils.arg('--cpu',
           metavar='<cpu>',
           help='The number of virtual cpus.')
@utils.arg('-m', '--memory',
           metavar='<memory>',
           help='The container memory size in MiB')
@utils.arg('--name',
           metavar='<name>',
           help='The new name for the container')
@utils.exclusive_arg(
    'auto_heal_value',
    '--auto-heal',
    required=False, action='store_true',
    help='Automatic recovery the status of contaier')
@utils.exclusive_arg(
    'auto_heal_value',
    '--no-auto-heal',
    required=False, action='store_true',
    help='Needless recovery the status of contaier')
def do_update(cs, args):
    """Update one or more attributes of the container."""
    opts = {}
    opts['memory'] = args.memory
    opts['cpu'] = args.cpu
    opts['name'] = args.name
    if 'auto_heal' in args and args.auto_heal:
        opts['auto_heal'] = True
    if 'no_auto_heal' in args and args.no_auto_heal:
        opts['auto_heal'] = False
    opts = zun_utils.remove_null_parms(**opts)
    if not opts:
        raise exc.CommandError("You must update at least one property")
    container = cs.containers.update(args.container, **opts)
    _show_container(container)


@utils.deprecated(RENAME_DEPRECATION_MESSAGE)
@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to rename.')
@utils.arg('name',
           metavar='<name>',
           help='The new name for the container')
def do_rename(cs, args):
    """Rename a container."""
    cs.containers.rename(args.container, args.name)


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to be attached to.')
def do_attach(cs, args):
    """Attach to a running container."""
    response = cs.containers.attach(args.container)
    websocketclient.do_attach(cs, response, args.container, "~", 0.5)


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to display processes.')
@utils.arg('--pid',
           metavar='<pid>',
           action='append', default=[],
           help='The args of the ps id.')
def do_top(cs, args):
    """Display the running processes inside the container."""

    if args.pid:
        # List container single ps id top result
        output = cs.containers.top(args.container, ' '.join(args.pid))
    else:
        # List container all processes top result
        output = cs.containers.top(args.container)
    for titles in output['Titles']:
        print("%-20s") % titles,
    if output['Processes']:
        for process in output['Processes']:
            print("")
            for info in process:
                print("%-20s") % info,
    else:
        print("")


@utils.arg('source',
           metavar='<source>',
           help='The source should be copied to the container or localhost. '
                'The format of this parameter is [container:]src_path.')
@utils.arg('destination',
           metavar='<destination>',
           help='The directory destination where save the source. '
                'The format of this parameter is [container:]dest_path.')
def do_cp(cs, args):
    """Copy files/tars between a container and the local filesystem."""
    if ':' in args.source:
        source_parts = args.source.split(':', 1)
        container_id = source_parts[0]
        container_path = source_parts[1]
        opts = {}
        opts['id'] = container_id
        opts['path'] = container_path

        res = cs.containers.get_archive(**opts)
        dest_path = args.destination
        tardata = io.BytesIO(res['data'])
        with closing(tarfile.open(fileobj=tardata)) as tar:
            tar.extractall(dest_path)  # nosec

    elif ':' in args.destination:
        dest_parts = args.destination.split(':', 1)
        container_id = dest_parts[0]
        container_path = dest_parts[1]
        filename = os.path.split(args.source)[1]
        opts = {}
        opts['id'] = container_id
        opts['path'] = container_path
        tardata = io.BytesIO()
        with closing(tarfile.open(fileobj=tardata, mode='w')) as tar:
            tar.add(args.source, arcname=filename)
        opts['data'] = tardata.getvalue()
        cs.containers.put_archive(**opts)

    else:
        print("Please check the parameters for zun copy!")
        print("Usage:")
        print("zun cp container:src_path dest_path|-")
        print("zun cp src_path|- container:dest_path")


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to display stats.')
def do_stats(cs, args):
    """Display stats of the container."""
    stats_info = cs.containers.stats(args.container)
    utils.print_dict(stats_info)


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to commit.')
@utils.arg('repository',
           metavar='<repository>[:<tag>]',
           help='The repository and tag of the image.')
def do_commit(cs, args):
    """Create a new image from a container's changes."""
    opts = zun_utils.check_commit_container_args(args)
    opts = zun_utils.remove_null_parms(**opts)
    try:
        image = cs.containers.commit(args.container, **opts)
        print("Request to commit container %s has been accepted. "
              "The image is %s." % (args.container, image['uuid']))
    except Exception as e:
        print("Commit for container %(container)s failed: %(e)s" %
              {'container': args.container, 'e': e})


@utils.deprecated(SG_DEPRECATION_MESSAGE)
@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to add security group.')
@utils.arg('security_group',
           metavar='<security_group>',
           help='Security group ID or name for specified container.')
def do_add_security_group(cs, args):
    """Add security group for specified container."""
    opts = {}
    opts['id'] = args.container
    opts['security_group'] = args.security_group
    opts = zun_utils.remove_null_parms(**opts)
    try:
        cs.containers.add_security_group(**opts)
        print("Request to add security group for container %s "
              "has been accepted." % args.container)
    except Exception as e:
        print("Add security group for container %(container)s "
              "failed: %(e)s" % {'container': args.container, 'e': e})


@utils.exclusive_arg(
    'detach_network_port',
    '--network',
    metavar='<network>',
    help='The neutron network that container will detach from.')
@utils.exclusive_arg(
    'detach_network_port',
    '--port',
    metavar='<port>',
    help='The neutron port that container will detach from.')
@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to detach the network.')
def do_network_detach(cs, args):
    """Detach a network from the container."""
    opts = {}
    opts['container'] = args.container
    opts['network'] = args.network
    opts['port'] = args.port
    opts = zun_utils.remove_null_parms(**opts)
    try:
        cs.containers.network_detach(**opts)
        print("Request to detach network from container %s "
              "has been accepted." % args.container)
    except Exception as e:
        print("Detach network from container %(container)s "
              "failed: %(e)s" % {'container': args.container, 'e': e})


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to attach network.')
@utils.arg('--network',
           metavar='<network>',
           help='The neutron network that container will attach to.')
@utils.arg('--port',
           metavar='<port>',
           help='The neutron port that container will attach to.')
@utils.arg('--fixed-ip',
           metavar='<fixed_ip>',
           help='The fixed-ip that container will attach to.')
def do_network_attach(cs, args):
    """Attach a network to the container."""
    opts = {}
    opts['container'] = args.container
    opts['network'] = args.network
    opts['port'] = args.port
    opts['fixed_ip'] = args.fixed_ip
    opts = zun_utils.remove_null_parms(**opts)
    try:
        cs.containers.network_attach(**opts)
        print("Request to attach network to container %s "
              "has been accepted." % args.container)
    except Exception as e:
        print("Attach network to container %(container)s "
              "failed: %(e)s" % {'container': args.container, 'e': e})


@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to display network info.')
def do_network_list(cs, args):
    """List networks on a container"""
    opts = {}
    opts['container'] = args.container
    opts = zun_utils.remove_null_parms(**opts)
    networks = cs.containers.network_list(**opts)
    zun_utils.list_container_networks(networks)


@utils.deprecated(SG_DEPRECATION_MESSAGE)
@utils.arg('container',
           metavar='<container>',
           help='ID or name of the container to remove security group.')
@utils.arg('security_group',
           metavar='<security_group>',
           help='The security group to remove from specified container.')
def do_remove_security_group(cs, args):
    """Remove security group for specified container."""
    opts = {}
    opts['id'] = args.container
    opts['security_group'] = args.security_group
    opts = zun_utils.remove_null_parms(**opts)
    try:
        cs.containers.remove_security_group(**opts)
        print("Request to remove security group for container %s "
              "has been accepted." % args.container)
    except Exception as e:
        print("Remove security group for container %(container)s "
              "failed: %(e)s" % {'container': args.container, 'e': e})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import testtools
from testtools import matchers
from urllib import parse
from zunclient.common import utils as zun_utils
from zunclient import exceptions
from zunclient.tests.unit import utils
from zunclient.v1 import containers

CONTAINER1 = {'id': '1234',
              'uuid': '36e527e4-6d03-4eda-9443-904424043741',
              'name': 'test1',
              'image_pull_policy': 'never',
              'image': 'cirros',
              'command': 'sh -c "echo hello"',
              'cpu': '1',
              'memory': '256',
              'environment': {'hostname': 'zunsystem'},
              'workdir': '/',
              'labels': {'label1': 'foo'},
              'hints': {'hint1': 'bar'},
              'restart_policy': 'no',
              'security_groups': ['test'],
              'auto_remove': True,
              'runtime': 'runc',
              'hostname': 'testhost',
              'disk': '20',
              'auto_heal': False,
              'privileged': False,
              'healthcheck': {}
              }

CONTAINER2 = {'id': '1235',
              'uuid': 'c7f9da0f-581b-4586-8d0d-a6c894822165',
              'name': 'test2',
              'image_pull_policy': 'ifnotpresent',
              'image': 'cirros',
              'command': 'sleep 100000000',
              'cpu': '1',
              'memory': '256',
              'environment': {'hostname': 'zunsystem'},
              'workdir': '/',
              'labels': {'label2': 'foo'},
              'hints': {'hint2': 'bar'},
              'restart_policy': 'on-failure:5',
              'security_groups': ['test'],
              'auto_remove': False,
              'runtime': 'runc',
              'hostname': 'testhost',
              'auto_heal': False,
              'privileged': True,
              'healthcheck': {}
              }

NETWORK1 = {'net_id': '99e90853-e1fd-4c57-a116-9e335deaa592',
            'port_id': '83f39a10-45c8-4463-a274-5a7cda3e6c97',
            'fixed_ips': [{
                'ip_address': '10.0.0.7', 'version': 4,
                'subnet_id': '5899aa85-c98f-4d1d-bc8f-99fed7bde5b9'}]
            }


CREATE_CONTAINER1 = copy.deepcopy(CONTAINER1)
del CREATE_CONTAINER1['id']
del CREATE_CONTAINER1['uuid']

force_delete1 = False
force_delete2 = True
all_projects = True
signal = "SIGTERM"
name = "new-name"
timeout = 10
tty_height = "56"
tty_width = "121"
path = "/tmp/test.txt"
data = "/tmp/test.tar"
repo = "repo-test"
tag = "tag-test"
security_group = "testsg"

fake_responses = {
    '/v1/containers':
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
        'POST': (
            {},
            CREATE_CONTAINER1,
        ),
    },
    '/v1/containers/?limit=2':
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
    },
    '/v1/containers/?marker=%s' % CONTAINER2['uuid']:
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
    },
    '/v1/containers/?limit=2&marker=%s' % CONTAINER2['uuid']:
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
    },
    '/v1/containers/?sort_dir=asc':
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
    },
    '/v1/containers/?sort_key=uuid':
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
    },
    '/v1/containers/?sort_key=uuid&sort_dir=desc':
    {
        'GET': (
            {},
            {'containers': [CONTAINER1, CONTAINER2]},
        ),
    },
    '/v1/containers/%s' % CONTAINER1['id']:
    {
        'GET': (
            {},
            CONTAINER1
        ),
    },
    '/v1/containers/%s' % CONTAINER1['name']:
    {
        'GET': (
            {},
            CONTAINER1
        ),
    },
    '/v1/containers/%s/start' % CONTAINER1['id']:
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s?force=%s' % (CONTAINER1['id'], force_delete1):
    {
        'DELETE': (
            {},
            None,
        ),
    },
    '/v1/containers/%s?force=%s' % (CONTAINER1['id'], force_delete2):
    {
        'DELETE': (
            {},
            None,
        ),
    },
    '/v1/containers/%s?all_projects=%s' % (CONTAINER1['id'], all_projects):
    {
        'DELETE': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/stop?timeout=10' % CONTAINER1['id']:
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/reboot?timeout=10' % CONTAINER1['id']:
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/pause' % CONTAINER1['id']:
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/unpause' % CONTAINER1['id']:
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/logs?%s'
    % (CONTAINER1['id'], parse.urlencode({'stdout': True, 'stderr': True,
                                          'timestamps': False, 'tail': 'all',
                                          'since': None})):
    {
        'GET': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/execute?%s'
    % (CONTAINER1['id'], parse.urlencode({'command': CONTAINER1['command']})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/kill?%s' % (CONTAINER1['id'],
                                   parse.urlencode({'signal': signal})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers?run=true':
    {
        'POST': (
            {},
            CREATE_CONTAINER1,
        ),
    },
    '/v1/containers/%s/rename?%s' % (CONTAINER1['id'],
                                     parse.urlencode({'name': name})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/attach' % CONTAINER1['id']:
    {
        'GET': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/resize?w=%s&h=%s'
    % (CONTAINER1['id'], tty_width, tty_height):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/resize?h=%s&w=%s'
    % (CONTAINER1['id'], tty_height, tty_width):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/top?ps_args=None' % (CONTAINER1['id']):
    {
        'GET': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/get_archive?%s'
    % (CONTAINER1['id'], parse.urlencode({'path': path})):
    {
        'GET': (
            {},
            {'data': data},
        ),
    },
    '/v1/containers/%s/put_archive?%s'
    % (CONTAINER1['id'], parse.urlencode({'path': path})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/stats?%s'
    % (CONTAINER1['id'], parse.urlencode({'decode': False,
                                          'stream': False})):
    {
        'GET': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/commit?%s'
    % (CONTAINER1['id'], parse.urlencode({'repository': repo,
                                          'tag': tag})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/add_security_group?%s'
    % (CONTAINER1['id'], parse.urlencode({'name': security_group})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/network_detach?%s'
    % (CONTAINER1['id'], parse.urlencode({'network': 'neutron_network'})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/network_attach?%s'
    % (CONTAINER1['id'], parse.urlencode({'network': 'neutron_network'})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/network_list'
    % (CONTAINER1['id']):
    {
        'GET': (
            {},
            {'networks': NETWORK1},
        ),
    },
    '/v1/containers/%s/remove_security_group?%s'
    % (CONTAINER1['id'], parse.urlencode({'name': security_group})):
    {
        'POST': (
            {},
            None,
        ),
    },
    '/v1/containers/%s/rebuild?image=cirros'
    % (CONTAINER1['id']):
    {
        'POST': (
            {},
            None,
        ),
    },
}


class ContainerManagerTest(testtools.TestCase):

    def setUp(self):
        super(ContainerManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = containers.ContainerManager(self.api)

    def test_container_create(self):
        containers = self.mgr.create(**CREATE_CONTAINER1)
        expect = [
            ('POST', '/v1/containers', {}, CREATE_CONTAINER1)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_container_create_fail(self):
        create_container_fail = copy.deepcopy(CREATE_CONTAINER1)
        create_container_fail["wrong_key"] = "wrong"
        self.assertRaisesRegex(exceptions.InvalidAttribute,
                               ("Key must be in %s" %
                                ','.join(containers.CREATION_ATTRIBUTES)),
                               self.mgr.create, **create_container_fail)
        self.assertEqual([], self.api.calls)

    def test_containers_list(self):
        containers = self.mgr.list()
        expect = [
            ('GET', '/v1/containers', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(containers, matchers.HasLength(2))

    def _test_containers_list_with_filters(self, limit=None, marker=None,
                                           sort_key=None, sort_dir=None,
                                           expect=[]):
        containers_filter = self.mgr.list(limit=limit, marker=marker,
                                          sort_key=sort_key,
                                          sort_dir=sort_dir)
        self.assertEqual(expect, self.api.calls)
        self.assertThat(containers_filter, matchers.HasLength(2))

    def test_containers_list_with_limit(self):
        expect = [
            ('GET', '/v1/containers/?limit=2', {}, None),
        ]
        self._test_containers_list_with_filters(
            limit=2,
            expect=expect)

    def test_containers_list_with_marker(self):
        expect = [
            ('GET', '/v1/containers/?marker=%s' % CONTAINER2['uuid'],
             {}, None),
        ]
        self._test_containers_list_with_filters(
            marker=CONTAINER2['uuid'],
            expect=expect)

    def test_containers_list_with_marker_limit(self):
        expect = [
            ('GET', '/v1/containers/?limit=2&marker=%s' % CONTAINER2['uuid'],
             {}, None),
        ]
        self._test_containers_list_with_filters(
            limit=2, marker=CONTAINER2['uuid'],
            expect=expect)

    def test_coontainer_list_with_sort_dir(self):
        expect = [
            ('GET', '/v1/containers/?sort_dir=asc', {}, None),
        ]
        self._test_containers_list_with_filters(
            sort_dir='asc',
            expect=expect)

    def test_container_list_with_sort_key(self):
        expect = [
            ('GET', '/v1/containers/?sort_key=uuid', {}, None),
        ]
        self._test_containers_list_with_filters(
            sort_key='uuid',
            expect=expect)

    def test_container_list_with_sort_key_dir(self):
        expect = [
            ('GET', '/v1/containers/?sort_key=uuid&sort_dir=desc', {}, None),
        ]
        self._test_containers_list_with_filters(
            sort_key='uuid', sort_dir='desc',
            expect=expect)

    def test_container_show_by_id(self):
        container = self.mgr.get(CONTAINER1['id'])
        expect = [
            ('GET', '/v1/containers/%s' % CONTAINER1['id'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(CONTAINER1['name'], container.name)
        self.assertEqual(CONTAINER1['uuid'], container.uuid)

    def test_container_show_by_name(self):
        container = self.mgr.get(CONTAINER1['name'])
        expect = [
            ('GET', '/v1/containers/%s' % CONTAINER1['name'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(CONTAINER1['name'], container.name)
        self.assertEqual(CONTAINER1['uuid'], container.uuid)

    def test_containers_start(self):
        containers = self.mgr.start(CONTAINER1['id'])
        expect = [
            ('POST', '/v1/containers/%s/start' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_delete(self):
        containers = self.mgr.delete(CONTAINER1['id'], force=force_delete1)
        expect = [
            ('DELETE', '/v1/containers/%s?force=%s' % (CONTAINER1['id'],
                                                       force_delete1),
             {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_delete_with_force(self):
        containers = self.mgr.delete(CONTAINER1['id'], force=force_delete2)
        expect = [
            ('DELETE', '/v1/containers/%s?force=%s' % (CONTAINER1['id'],
                                                       force_delete2),
             {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_delete_with_all_projects(self):
        containers = self.mgr.delete(CONTAINER1['id'],
                                     all_projects=all_projects)
        expect = [
            ('DELETE', '/v1/containers/%s?all_projects=%s' % (CONTAINER1['id'],
                                                              all_projects),
             {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_stop(self):
        containers = self.mgr.stop(CONTAINER1['id'], timeout)
        expect = [
            ('POST', '/v1/containers/%s/stop?timeout=10' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_restart(self):
        containers = self.mgr.restart(CONTAINER1['id'], timeout)
        expect = [
            ('POST', '/v1/containers/%s/reboot?timeout=10' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_pause(self):
        containers = self.mgr.pause(CONTAINER1['id'])
        expect = [
            ('POST', '/v1/containers/%s/pause' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_unpause(self):
        containers = self.mgr.unpause(CONTAINER1['id'])
        expect = [
            ('POST', '/v1/containers/%s/unpause' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_logs(self):
        containers = self.mgr.logs(CONTAINER1['id'], stdout=True, stderr=True,
                                   timestamps=False, tail='all', since=None)
        expect = [
            ('GET', '/v1/containers/%s/logs?%s'
             % (CONTAINER1['id'], parse.urlencode({'stdout': True,
                                                   'stderr': True,
                                                   'timestamps': False,
                                                   'tail': 'all',
                                                   'since': None})),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_execute(self):
        containers = self.mgr.execute(CONTAINER1['id'],
                                      command=CONTAINER1['command'])
        expect = [
            ('POST', '/v1/containers/%s/execute?%s'
             % (CONTAINER1['id'], parse.urlencode({'command':
                                                   CONTAINER1['command']})),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_kill(self):
        containers = self.mgr.kill(CONTAINER1['id'], signal)
        expect = [
            ('POST', '/v1/containers/%s/kill?%s'
             % (CONTAINER1['id'], parse.urlencode({'signal': signal})),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_container_run(self):
        containers = self.mgr.run(**CREATE_CONTAINER1)
        expect = [
            ('POST', '/v1/containers?run=true', {}, CREATE_CONTAINER1)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_container_run_fail(self):
        run_container_fail = copy.deepcopy(CREATE_CONTAINER1)
        run_container_fail["wrong_key"] = "wrong"
        self.assertRaisesRegex(exceptions.InvalidAttribute,
                               ("Key must be in %s" %
                                ','.join(containers.CREATION_ATTRIBUTES)),
                               self.mgr.run, **run_container_fail)
        self.assertEqual([], self.api.calls)

    def test_containers_rename(self):
        containers = self.mgr.rename(CONTAINER1['id'], name)
        expect = [
            ('POST', '/v1/containers/%s/rename?%s'
             % (CONTAINER1['id'], parse.urlencode({'name': name})),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_attach(self):
        containers = self.mgr.attach(CONTAINER1['id'])
        expect = [
            ('GET', '/v1/containers/%s/attach' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_resize(self):
        containers = self.mgr.resize(CONTAINER1['id'], tty_width, tty_height)
        expects = []
        expects.append([
            ('POST', '/v1/containers/%s/resize?w=%s&h=%s'
             % (CONTAINER1['id'], tty_width, tty_height),
             {'Content-Length': '0'}, None)
        ])
        expects.append([
            ('POST', '/v1/containers/%s/resize?h=%s&w=%s'
             % (CONTAINER1['id'], tty_height, tty_width),
             {'Content-Length': '0'}, None)
        ])
        self.assertTrue(self.api.calls in expects)
        self.assertIsNone(containers)

    def test_containers_top(self):
        containers = self.mgr.top(CONTAINER1['id'])
        expect = [
            ('GET', '/v1/containers/%s/top?ps_args=None' % CONTAINER1['id'],
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_get_archive(self):
        response = self.mgr.get_archive(CONTAINER1['id'], path)
        expect = [
            ('GET', '/v1/containers/%s/get_archive?%s'
             % (CONTAINER1['id'], parse.urlencode({'path': path})),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(zun_utils.decode_file_data(data), response['data'])

    def test_containers_put_archive(self):
        response = self.mgr.put_archive(CONTAINER1['id'], path, data)
        expect = [
            ('POST', '/v1/containers/%s/put_archive?%s'
             % (CONTAINER1['id'], parse.urlencode({'path': path})),
             {'Content-Length': '0'},
             {'data': zun_utils.encode_file_data(data)})
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(response)

    def test_containers_commit(self):
        containers = self.mgr.commit(CONTAINER1['id'], repo, tag)
        expect = [
            ('POST', '/v1/containers/%s/commit?%s' % (CONTAINER1['id'],
             parse.urlencode({'repository': repo, 'tag': tag})),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(containers)

    def test_containers_add_security_group(self):
        containers = self.mgr.add_security_group(
            CONTAINER1['id'], security_group)
        expect = [
            ('POST', '/v1/containers/%s/add_security_group?%s'
             % (CONTAINER1['id'],
                parse.urlencode({'name': security_group})),
                {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_network_detach(self):
        containers = self.mgr.network_detach(
            CONTAINER1['id'], network='neutron_network')
        expect = [
            ('POST', '/v1/containers/%s/network_detach?%s'
             % (CONTAINER1['id'],
                parse.urlencode({'network': 'neutron_network'})),
                {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_network_attach(self):
        containers = self.mgr.network_attach(
            CONTAINER1['id'], network='neutron_network')
        expect = [
            ('POST', '/v1/containers/%s/network_attach?%s'
             % (CONTAINER1['id'],
                parse.urlencode({'network': 'neutron_network'})),
                {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_network_list(self):
        networks = self.mgr.network_list(CONTAINER1['id'])
        expect = [
            ('GET', '/v1/containers/%s/network_list'
             % (CONTAINER1['id']), {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(networks)

    def test_containers_remove_security_group(self):
        containers = self.mgr.remove_security_group(
            CONTAINER1['id'], security_group)
        expect = [
            ('POST', '/v1/containers/%s/remove_security_group?%s'
             % (CONTAINER1['id'],
                parse.urlencode({'name': security_group})),
                {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(containers)

    def test_containers_rebuild(self):
        opts = {'image': 'cirros'}
        self.mgr.rebuild(CONTAINER1['id'], **opts)
        expect = [
            ('POST',
             '/v1/containers/%s/rebuild?image=cirros'
             % (CONTAINER1['id']),
             {'Content-Length': '0'}, None)
        ]
        self.assertEqual(expect, self.api.calls)



#
# Copyright 2013 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from zunclient.common import cliutils
from zunclient.common import utils
from zunclient import exceptions as exc
from zunclient.tests.unit import utils as test_utils


class CommonFiltersTest(test_utils.BaseTestCase):
    def test_limit(self):
        result = utils.common_filters(limit=42)
        self.assertEqual(['limit=42'], result)

    def test_limit_0(self):
        result = utils.common_filters(limit=0)
        self.assertEqual(['limit=0'], result)

    def test_limit_negative_number(self):
        result = utils.common_filters(limit=-2)
        self.assertEqual(['limit=-2'], result)

    def test_other(self):
        for key in ('marker', 'sort_key', 'sort_dir'):
            result = utils.common_filters(**{key: 'test'})
            self.assertEqual(['%s=test' % key], result)


class SplitAndDeserializeTest(test_utils.BaseTestCase):

    def test_split_and_deserialize(self):
        ret = utils.split_and_deserialize('str=foo')
        self.assertEqual(('str', 'foo'), ret)

        ret = utils.split_and_deserialize('int=1')
        self.assertEqual(('int', 1), ret)

        ret = utils.split_and_deserialize('bool=false')
        self.assertEqual(('bool', False), ret)

        ret = utils.split_and_deserialize('list=[1, "foo", 2]')
        self.assertEqual(('list', [1, "foo", 2]), ret)

        ret = utils.split_and_deserialize('dict={"foo": 1}')
        self.assertEqual(('dict', {"foo": 1}), ret)

        ret = utils.split_and_deserialize('str_int="1"')
        self.assertEqual(('str_int', "1"), ret)

    def test_split_and_deserialize_fail(self):
        self.assertRaises(exc.CommandError,
                          utils.split_and_deserialize, 'foo:bar')


class ArgsArrayToPatchTest(test_utils.BaseTestCase):

    def test_args_array_to_patch(self):
        my_args = {
            'attributes': ['str=foo', 'int=1', 'bool=true',
                           'list=[1, 2, 3]', 'dict={"foo": "bar"}'],
        }
        patch = utils.args_array_to_patch(my_args['attributes'])
        self.assertEqual([{'str': 'foo'},
                          {'int': 1},
                          {'bool': True},
                          {'list': [1, 2, 3]},
                          {'dict': {"foo": "bar"}}], patch)


class FormatArgsTest(test_utils.BaseTestCase):

    def test_format_args_none(self):
        self.assertEqual({}, utils.format_args(None))

    def test_format_args(self):
        li = utils.format_args([
            'K1=V1,K2=V2,'
            'K3=V3,K4=V4,'
            'K5=V5'])
        self.assertEqual({'K1': 'V1',
                          'K2': 'V2',
                          'K3': 'V3',
                          'K4': 'V4',
                          'K5': 'V5'
                          }, li)

    def test_format_args_semicolon(self):
        li = utils.format_args([
            'K1=V1;K2=V2;'
            'K3=V3;K4=V4;'
            'K5=V5'])
        self.assertEqual({'K1': 'V1',
                          'K2': 'V2',
                          'K3': 'V3',
                          'K4': 'V4',
                          'K5': 'V5'
                          }, li)

    def test_format_args_mix_commas_semicolon(self):
        li = utils.format_args([
            'K1=V1,K2=V2,'
            'K3=V3;K4=V4,'
            'K5=V5'])
        self.assertEqual({'K1': 'V1',
                          'K2': 'V2',
                          'K3': 'V3',
                          'K4': 'V4',
                          'K5': 'V5'
                          }, li)

    def test_format_args_split(self):
        li = utils.format_args([
            'K1=V1,'
            'K2=V22222222222222222222222222222'
            '222222222222222222222222222,'
            'K3=3.3.3.3'])
        self.assertEqual({'K1': 'V1',
                          'K2': 'V22222222222222222222222222222'
                          '222222222222222222222222222',
                          'K3': '3.3.3.3'}, li)

    def test_format_args_multiple(self):
        li = utils.format_args([
            'K1=V1',
            'K2=V22222222222222222222222222222'
            '222222222222222222222222222',
            'K3=3.3.3.3'])
        self.assertEqual({'K1': 'V1',
                          'K2': 'V22222222222222222222222222222'
                          '222222222222222222222222222',
                          'K3': '3.3.3.3'}, li)

    def test_format_args_multiple_colon_values(self):
        li = utils.format_args([
            'K1=V1',
            'K2=V2,V22,V222,V2222',
            'K3=3.3.3.3'])
        self.assertEqual({'K1': 'V1',
                          'K2': 'V2,V22,V222,V2222',
                          'K3': '3.3.3.3'}, li)

    def test_format_args_parse_comma_false(self):
        li = utils.format_args(
            ['K1=V1,K2=2.2.2.2,K=V'],
            parse_comma=False)
        self.assertEqual({'K1': 'V1,K2=2.2.2.2,K=V'}, li)

    def test_format_args_multiple_values_per_args(self):
        li = utils.format_args([
            'K1=V1',
            'K1=V2'])
        self.assertIn('K1', li)
        self.assertIn('V1', li['K1'])
        self.assertIn('V2', li['K1'])

    def test_format_args_bad_arg(self):
        args = ['K1=V1,K22.2.2.2']
        ex = self.assertRaises(exc.CommandError,
                               utils.format_args, args)
        self.assertEqual('arguments must be a list of KEY=VALUE '
                         'not K22.2.2.2', str(ex))

    def test_format_multiple_bad_args(self):
        args = ['K1=V1', 'K22.2.2.2']
        ex = self.assertRaises(exc.CommandError,
                               utils.format_args, args)
        self.assertEqual('arguments must be a list of KEY=VALUE '
                         'not K22.2.2.2', str(ex))


class CliUtilsTest(test_utils.BaseTestCase):

    def test_keys_and_vals_to_strs(self):
        dict_in = {'a': '1',
                   'b': {
                       'x': 1,
                       'y': '2',
                       'z': '3'},
                   'c': 7}

        dict_exp = collections.OrderedDict([
            ('a', '1'),
            ('b', collections.OrderedDict([
                ('x', 1),
                ('y', '2'),
                ('z', '3')])),
            ('c', 7)])

        dict_out = cliutils.keys_and_vals_to_strs(dict_in)
        dict_act = collections.OrderedDict([
            ('a', dict_out['a']),
            ('b', collections.OrderedDict(sorted(dict_out['b'].items()))),
            ('c', dict_out['c'])])

        self.assertEqual(str(dict_exp), str(dict_act))


class ParseNetsTest(test_utils.BaseTestCase):

    def test_no_nets(self):
        nets = []
        result = utils.parse_nets(nets)
        self.assertEqual([], result)

    def test_nets_with_network(self):
        nets = [' network = 1234567 , v4-fixed-ip = 172.17.0.3 ']
        result = utils.parse_nets(nets)
        self.assertEqual([{'network': '1234567', 'v4-fixed-ip': '172.17.0.3'}],
                         result)

    def test_nets_with_port(self):
        nets = ['port=1234567, v6-fixed-ip=2001:db8::2']
        result = utils.parse_nets(nets)
        self.assertEqual([{'port': '1234567', 'v6-fixed-ip': '2001:db8::2'}],
                         result)

    def test_nets_with_only_ip(self):
        nets = ['v4-fixed-ip = 172.17.0.3']
        self.assertRaises(exc.CommandError,
                          utils.parse_nets, nets)

    def test_nets_with_both_network_port(self):
        nets = ['port=1234567, network=2345678, v4-fixed-ip=172.17.0.3']
        self.assertRaises(exc.CommandError,
                          utils.parse_nets, nets)

    def test_nets_with_invalid_ip(self):
        nets = ['network=1234567, v4-fixed-ip=23.555.567,789']
        self.assertRaises(exc.CommandError,
                          utils.parse_nets, nets)


class ParseCommandTest(test_utils.BaseTestCase):

    def test_no_command(self):
        command = []
        result = utils.parse_command(command)
        self.assertEqual('', result)

    def test_command_ls(self):
        command = ['ls', '-al']
        result = utils.parse_command(command)
        self.assertEqual('"ls" "-al"', result)

    def test_command_echo_hello(self):
        command = ['sh', '-c', 'echo hello']
        result = utils.parse_command(command)
        self.assertEqual('"sh" "-c" "echo hello"', result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from oslo_serialization import jsonutils

from zunclient.common import cliutils as utils
from zunclient.common import utils as zun_utils
from zunclient import exceptions as exc


def _show_registry(registry):
    utils.print_dict(registry._info['registry'])


@utils.arg('--name',
           metavar='<name>',
           help='The name of the registry.')
@utils.arg('--username',
           metavar='<username>',
           help='The username to login to the registry.')
@utils.arg('--password',
           metavar='<password>',
           help='The password to login to the registry.')
@utils.arg('--domain',
           metavar='<domain>',
           required=True,
           help='The domain of the registry.')
def do_registry_create(cs, args):
    """Create a registry."""
    opts = {}
    opts['name'] = args.name
    opts['domain'] = args.domain
    opts['username'] = args.username
    opts['password'] = args.password
    opts = zun_utils.remove_null_parms(**opts)
    _show_registry(cs.registries.create(**opts))


@utils.arg('--all-projects',
           action="store_true",
           default=False,
           help='List registries in all projects')
@utils.arg('--marker',
           metavar='<marker>',
           default=None,
           help='The last registry UUID of the previous page; '
                'displays list of registries after "marker".')
@utils.arg('--limit',
           metavar='<limit>',
           type=int,
           help='Maximum number of registries to return')
@utils.arg('--sort-key',
           metavar='<sort-key>',
           help='Column to sort results by')
@utils.arg('--sort-dir',
           metavar='<sort-dir>',
           choices=['desc', 'asc'],
           help='Direction to sort. "asc" or "desc".')
@utils.arg('--name',
           metavar='<name>',
           help='List registries according to their name.')
@utils.arg('--domain',
           metavar='<domain>',
           help='List registries according to their domain.')
@utils.arg('--project-id',
           metavar='<project-id>',
           help='List registries according to their Project_id')
@utils.arg('--user-id',
           metavar='<user-id>',
           help='List registries according to their user_id')
@utils.arg('--username',
           metavar='<username>',
           help='List registries according to their username')
def do_registry_list(cs, args):
    """Print a list of available registries."""
    opts = {}
    opts['all_projects'] = args.all_projects
    opts['marker'] = args.marker
    opts['limit'] = args.limit
    opts['sort_key'] = args.sort_key
    opts['sort_dir'] = args.sort_dir
    opts['domain'] = args.domain
    opts['name'] = args.name
    opts['project_id'] = args.project_id
    opts['user_id'] = args.user_id
    opts['username'] = args.username
    opts = zun_utils.remove_null_parms(**opts)
    registries = cs.registries.list(**opts)
    columns = ('uuid', 'name', 'domain', 'username', 'password')
    utils.print_list(registries, columns,
                     sortby_index=None)


@utils.arg('registries',
           metavar='<registry>',
           nargs='+',
           help='ID or name of the (registry)s to delete.')
def do_registry_delete(cs, args):
    """Delete specified registries."""
    for registry in args.registries:
        opts = {}
        opts['id'] = registry
        opts = zun_utils.remove_null_parms(**opts)
        try:
            cs.registries.delete(**opts)
            print("Request to delete registry %s has been accepted." %
                  registry)
        except Exception as e:
            print("Delete for registry %(registry)s failed: %(e)s" %
                  {'registry': registry, 'e': e})


@utils.arg('registry',
           metavar='<registry>',
           help='ID or name of the registry to show.')
@utils.arg('-f', '--format',
           metavar='<format>',
           action='store',
           choices=['json', 'yaml', 'table'],
           default='table',
           help='Print representation of the container.'
                'The choices of the output format is json,table,yaml.'
                'Defaults to table.')
def do_registry_show(cs, args):
    """Show details of a registry."""
    opts = {}
    opts['id'] = args.registry
    opts = zun_utils.remove_null_parms(**opts)
    registry = cs.registries.get(**opts)
    if args.format == 'json':
        print(jsonutils.dumps(registry._info, indent=4, sort_keys=True))
    elif args.format == 'yaml':
        print(yaml.safe_dump(registry._info, default_flow_style=False))
    elif args.format == 'table':
        _show_registry(registry)


@utils.arg('registry',
           metavar='<registry>',
           help="ID or name of the registry to update.")
@utils.arg('--username',
           metavar='<username>',
           help='The new username for the registry.')
@utils.arg('--password',
           metavar='<password>',
           help='The new password for the registry.')
@utils.arg('--domain',
           metavar='<domain>',
           help='The new domain for the registry.')
@utils.arg('--name',
           metavar='<name>',
           help='The new name for the registry')
def do_registry_update(cs, args):
    """Update one or more attributes of the registry."""
    opts = {}
    opts['username'] = args.username
    opts['password'] = args.password
    opts['domain'] = args.domain
    opts['name'] = args.name
    opts = zun_utils.remove_null_parms(**opts)
    if not opts:
        raise exc.CommandError("You must update at least one property")
    registry = cs.registries.update(args.registry, **opts)
    _show_registry(registry)
